# TODO: Add License
import copy
import json
import pathlib
import random
import shutil
import tempfile
import textwrap
import time
from collections import namedtuple
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Union

import numpy as np
import pandas as pd
import yaml
from deepdiff import DeepDiff
from deprecated import deprecated
from werkzeug.datastructures import FileStorage

from fiddler.api.monitoring import Monitoring
from fiddler.api.publish_event import PublishEvent
from fiddler.api.segment import Segment
from fiddler.connection import Connection
from fiddler.experimental import ExperimentalFeatures
from fiddler.file_processor.src.constants import (
    CSV_EXTENSION,
    PARQUET_EXTENSION,
    SUPPORTABLE_FILE_EXTENSIONS,
)
from fiddler.project import Project
from fiddler.utils import logging

from . import constants
from .core_objects import (
    AttributionExplanation,
    BatchPublishType,
    DatasetInfo,
    DataType,
    DeploymentOptions,
    DeploymentType,
    FiddlerTimestamp,
    MLFlowParams,
    ModelInfo,
    ModelTask,
    MonitoringViolation,
    MonitoringViolationType,
    MulticlassAttributionExplanation,
    SegmentInfo,
)
from .file_processor.src.facade import upload_dataset
from .model_info_validator import ModelInfoValidator
from .monitoring_validator import MonitoringValidator
from .utils import cast_input_data
from .utils.formatting import model_string_to_hostname, print_streamed_result
from .utils.general_checks import do_not_proceed, safe_name_check, type_enforce
from .utils.pandas import (
    df_size_exceeds,
    try_series_retype,
    write_dataframe_to_parquet_file,
)

LOG = logging.getLogger(__name__)

SUCCESS_STATUS = Connection.SUCCESS_STATUS
FAILURE_STATUS = Connection.FAILURE_STATUS
FIDDLER_ARGS_KEY = Connection.FIDDLER_ARGS_KEY
STREAMING_HEADER_KEY = Connection.STREAMING_HEADER_KEY
AUTH_HEADER_KEY = Connection.AUTH_HEADER_KEY
ROUTING_HEADER_KEY = Connection.ROUTING_HEADER_KEY
ADMIN_SERVICE_PORT = 4100
DATASET_MAX_ROWS = 50_000

# A PredictionEventBundle represents a batch of inferences and their input
# features. All of these share schema, latency, and success status. A bundle
# can consist just one event as well.
PredictionEventBundle = namedtuple(
    'PredictionEventBundle',
    [
        'prediction_status',  # typeof: int # 0 for success, failure otherwise
        'prediction_latency',  # typeof: float # Latency in seconds.
        'input_feature_bundle',  # list of feature vectors.
        'prediction_bundle',  # list of prediction vectors.
        # TODO: Support sending schema as well.
    ],
)

_protocol_version = 1


class FiddlerApi:
    """Broker of all connections to the Fiddler API.
    Conventions:
        - Exceptions are raised for FAILURE reponses from the backend.
        - Methods beginning with `list` fetch lists of ids (e.g. all model ids
            for a project) and do not alter any data or state on the backend.
        - Methods beginning with `get` return a more complex object but also
            do not alter data or state on the backend.
        - Methods beginning with `run` invoke model-related execution and
            return the result of that computation. These do not alter state,
            but they can put a heavy load on the computational resources of
            the Fiddler engine, so they should be paralellized with care.
        - Methods beginning with `delete` permanently, irrevocably, and
            globally destroy objects in the backend. Think "rm -rf"
        - Methods beginning with `upload` convert and transmit supported local
            objects to Fiddler-friendly formats loaded into the Fiddler engine.
            Attempting to upload an object with an identifier that is already
            in use will result in an exception being raised, rather than the
            existing object being overwritten. To update an object in the
            Fiddler engine, please call both the `delete` and `upload` methods
            pertaining to the object in question.

    :param url: The base URL of the API to connect to. Usually either
        https://dev.fiddler.ai (cloud) or http://localhost:4100 (onebox)
    :param org_id: The name of your organization in the Fiddler platform
    :param auth_token: Token used to authenticate. Your token can be
        created, found, and changed at <FIDDLER URL>/settings/credentials.
    :param proxies: optionally, a dict of proxy URLs. e.g.,
                    proxies = {'http' : 'http://proxy.example.com:1234',
                               'https': 'https://proxy.example.com:5678'}
    :param verbose: if True, api calls will be logged verbosely,
                    *warning: all information required for debugging will be
                    logged including the auth_token.
    :param verify: if False, certificate verification will be disabled when
        establishing an SSL connection.
    """

    def __init__(
        self,
        url=None,
        org_id=None,
        auth_token=None,
        proxies=None,
        verbose=False,
        timeout: int = None,
        verify=True,
    ):
        self.org_id = org_id
        self.strict_mode = True
        self.connection = Connection(
            url, 
            org_id, 
            auth_token, 
            proxies, 
            verbose, 
            timeout,
            verify=verify
        )

        self.monitoring_validator = MonitoringValidator()
        self.experimental = ExperimentalFeatures(client=self)

    def __getattr__(self, function_name):
        """
        Overriding allows us to point unrecognized use cases to the documentation page
        """

        def method(*args, **kwargs):
            # This is a method that is not recognized
            error_msg = (
                f'Function `{function_name}` not found.\n'
                f'Please consult Fiddler documentation at `https://api.fiddler.ai/`'
            )
            raise RuntimeError(error_msg)

        return method

    @staticmethod
    def _abort_dataset_upload(
        dataset: Dict[str, pd.DataFrame], size_check_enabled: bool, max_len: int
    ):
        """
        This method checks if any of the dataframes exeeds size limit.
        In case the size limit is exceeded and size_check_enabled is True
        a warning is issued and the user is required to confirm if they'd
        like to proceed with the upload
        """
        # check if the dataset exceeds size limits
        warn_and_query = size_check_enabled and df_size_exceeds(dataset, max_len)
        if warn_and_query:
            LOG.warning(
                f'The dataset contains more than {max_len} datapoints. '
                f'Please allow for additional time to upload the dataset '
                f'and calculate statistical metrics. '
                f'To disable this message set the flag size_check_enabled to False. '
                f'\n\nAlternately, consider sampling the dataset. '
                f'If you plan to sample the dataset please ensure that the '
                f'representative sample captures all possible '
                f'categorical features, labels and numerical ranges that '
                f'would be encountered during deployment.'
                f'\n\nFor details on how datasets are used and considerations '
                f'for when large datasets are necessary, please refer to '
                f'https://docs.fiddler.ai/pages/user-guide/administration-concepts/project-structure/#dataset'
            )
            user_query = 'Would you like to proceed with the upload (y/n)? '
            return do_not_proceed(user_query)
        return False

    def _check_connection(self, check_client_version=True, check_server_version=True):
        return self.connection.check_connection(
            check_client_version, check_server_version
        )

    def _call_executor_service(
        self,
        path: List[str],
        json_payload: Any = None,
        files: Optional[List[Path]] = None,
        is_get_request: bool = False,
        stream: bool = False,
    ):
        return self.connection.call_executor_service(
            path, json_payload, files, is_get_request, stream
        )

    def _call(
        self,
        path: List[str],
        json_payload: Any = None,
        files: Optional[List[Path]] = None,
        is_get_request: bool = False,
        stream: bool = False,
        timeout: Optional[int] = None,
        num_tries: int = 1,
    ):
        """Issues a request to the API and returns the result,
        logigng and handling errors appropriately.

        Raises a RuntimeError if the response is a failure or cannot be parsed.
        Does not handle any ConnectionError exceptions thrown by the `requests`
        library.

        Note: Parameters `timeout` and `num_tries` are currently only utilized in a workaround
        for a bug involving Mac+Docker communication. See: https://github.com/docker/for-mac/issues/3448
        """
        return self.connection.call(
            path, json_payload, files, is_get_request, stream, timeout, num_tries
        )

    @deprecated(
        reason='Please use get_datasets, this method will be removed in future versions'
    )
    def list_datasets(self, project_id: str) -> List[str]:
        """List the ids of all datasets in the organization.

        :returns: List of strings containing the ids of each dataset.
        """
        return self.project(project_id).list_datasets()

    @deprecated(
        reason='Please use get_projects, this method will be removed in future versions'
    )
    def list_projects(self, get_project_details: bool = False) -> List[str]:
        """List the ids of all projects in the organization.

        :returns: List of strings containing the ids of each project.
        """
        path = ['list_projects', self.org_id]

        payload = {
            'project_details': get_project_details,
        }

        return self._call(path, json_payload=payload)['projects']

    def project(self, project_id):
        return Project(self.connection, project_id)

    @deprecated(
        reason='Please use get_models, this method will be removed in future versions'
    )
    def list_models(self, project_id: str) -> List[str]:
        """List the names of all models in a project.

        :param project_id: The unique identifier of the project on the Fiddler
            engine.
        :returns: List of strings containing the ids of each model in the
            specified project.
        """
        return self.project(project_id).list_models()

    @deprecated(
        reason='Please use get_dataset, this method will be removed in future versions'
    )
    def get_dataset_info(self, project_id: str, dataset_id: str) -> DatasetInfo:
        """Get DatasetInfo for a dataset.

        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine.

        :returns: A fiddler.DatasetInfo object describing the dataset.
        """
        return self.project(project_id).dataset(dataset_id).get_info()

    @deprecated(
        reason='Please use get_model, this method will be removed in future versions'
    )
    def get_model_info(self, project_id: str, model_id: str) -> ModelInfo:
        """Get ModelInfo for a model in a certain project.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.

        :returns: A fiddler.ModelInfo object describing the model.
        """
        return self.project(project_id).model(model_id).get_info()

    def _query_dataset(
        self,
        project_id: str,
        dataset_id: str,
        fields: List[str],
        max_rows: int,
        split: Optional[str] = None,
        sampling=False,
        sampling_seed=0.0,
    ):
        return (
            self.project(project_id)
            .dataset(dataset_id)
            ._query_dataset(fields, max_rows, split, sampling, sampling_seed)
        )

    @deprecated(
        reason='Please use get_slice, this method will be removed in future versions'
    )
    def get_dataset(
        self,
        project_id: str,
        dataset_id: str,
        max_rows: int = 1_000,
        splits: Optional[List[str]] = None,
        sampling=False,
        dataset_info: Optional[DatasetInfo] = None,
        include_fiddler_id=False,
    ) -> Dict[str, pd.DataFrame]:
        """Fetches data from a dataset on Fiddler.

        :param project_id: The unique identifier of the project on the Fiddler
            engine.
        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine.
        :param max_rows: Up to this many rows will be fetched from eash split
            of the dataset.
        :param splits: If specified, data will only be fetched for these
            splits. Otherwise, all splits will be fetched.
        :param sampling: If True, data will be sampled up to max_rows. If
            False, rows will be returned in order up to max_rows. The seed
            will be fixed for sampling.∂
        :param dataset_info: If provided, the API will skip looking up the
            DatasetInfo (a necessary precursor to requesting data).
        :param include_fiddler_id: Return the Fiddler engine internal id
            for each row. Useful only for debugging.

        :returns: A dictionary of str -> DataFrame that maps the name of
            dataset splits to the data in those splits. If len(splits) == 1,
            returns just that split as a dataframe, rather than a dataframe.
        """
        return (
            self.project(project_id)
            .dataset(dataset_id)
            .download(max_rows, splits, sampling, dataset_info, include_fiddler_id)
        )

    def get_slice(
        self,
        sql_query: str,
        project_id: str,
        columns_override: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """Fetches data from Fiddler via a *slice query* (SQL query).

        :param sql_query: A special SQL query that begins with the keyword
            "SLICE"
        :param project_id: The unique identifier of the project on the Fiddler
            engine.
        :param columns_override: A list of columns to return even if they are
            not specified in the slice.
        :returns: A table containing the sliced data (as a Pandas DataFrame)
        """
        payload: Dict[str, Any] = dict(sql=sql_query, project=project_id)
        if columns_override is not None:
            payload['slice_columns_override'] = columns_override

        path = ['slice_query', self.org_id, project_id]
        res = self._call(path, json_payload=payload)
        slice_info = res.pop(0)

        if not isinstance(slice_info, dict):
            raise RuntimeError(f'Invalid response from server - {slice_info}')

        if slice_info.get('status') == FAILURE_STATUS and slice_info.get('message'):
            raise Exception(slice_info.get('message'))

        if not slice_info.get('is_slice', False):
            raise RuntimeError(
                f'Query does not return a valid slice. info: {slice_info} Query: {sql_query}'
            )
        column_names = slice_info['columns']
        dtype_strings = slice_info['dtypes']
        df = pd.DataFrame(res, columns=column_names)
        for column_name, dtype in zip(column_names, dtype_strings):
            df[column_name] = try_series_retype(df[column_name], dtype)
        return df

    def delete_dataset(self, project_id: str, dataset_id: str):
        """Permanently delete a dataset.

        :param project_id: The unique identifier of the project on the Fiddler
            engine.
        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine.

        :returns: Server response for deletion action.
        """
        return self.project(project_id).dataset(dataset_id).delete()

    def delete_model(
        self, project_id: str, model_id: str, delete_prod=False, delete_pred=True
    ):
        """Permanently delete a model.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.
        :param delete_prod: Boolean value to delete the production table.
            By default this table is not dropped.
        :param delete_pred: Boolean value to delete the prediction table.
            By default this table is dropped.

        :returns: Server response for deletion action.
        """
        return self.project(project_id).model(model_id).delete(delete_prod, delete_pred)

    def _delete_model_artifacts(self, project_id: str, model_id: str):
        """Permanently delete a model artifacts.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.

        :returns: Server response for deletion action.
        """
        return self.project(project_id).model(model_id)._delete_artifacts()

    def delete_project(self, project_id: str):
        """Permanently delete a project.

        :param project_id: The unique identifier of the project on the Fiddler
            engine.

        :returns: Server response for deletion action.
        """
        # Type enforcement
        return self.project(project_id).delete()

    ##### Start: Methods related to uploading dataset #####

    def _upload_dataset_files(
        self,
        project_id: str,
        dataset_id: str,
        file_paths: List[Path],
        dataset_info: Optional[DatasetInfo] = None,
    ):
        """Uploads data files to the Fiddler platform."""
        safe_name_check(dataset_id, constants.MAX_ID_LEN, self.strict_mode)

        payload: Dict[str, Any] = dict(dataset_name=dataset_id)

        if dataset_info is not None:
            if self.strict_mode:
                dataset_info.validate()

            payload['dataset_info'] = dict(dataset=dataset_info.to_dict())

        payload['split_test'] = False
        path = ['dataset_upload', self.org_id, project_id]

        LOG.info(f'Uploading the dataset {dataset_id} ...')

        result = self._call(path, json_payload=payload, files=file_paths)

        return result

    def upload_dataset(
        self,
        project_id: str,
        dataset: Dict[str, pd.DataFrame],
        dataset_id: str,
        info: Optional[DatasetInfo] = None,
        size_check_enabled: bool = True,
    ):
        """Uploads a representative dataset to the Fiddler engine.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param dataset: A dictionary mapping name -> DataFrame
            containing data to be uploaded to the Fiddler engine.
        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine. Must be a short string without whitespace.
        :param info: A DatasetInfo object specifying all the details of the
            dataset. If not provided, a DatasetInfo will be inferred from the
            dataset and a warning raised.
        :param size_check_enabled: Flag to enable the dataframe size check.
            Default behavior is to raise a warning and present an interactive
            dialogue if the size of the dataframes exceeds the default limit.
            Set this flag to False to disable the checks.

        :returns: The server response for the upload.
        """
        # Type enforcement
        # @question: Shouldn't we just throw this as an error from server if we already don't do it?
        project_id = type_enforce('project_id', project_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)

        assert (
            ' ' not in dataset_id
        ), 'The dataset identifier should not contain whitespace'
        safe_name_check(dataset_id, constants.MAX_ID_LEN, self.strict_mode)

        # get a dictionary of str -> pd.DataFrame for all data to upload
        if not isinstance(dataset, dict):
            raise ValueError('dataset must be a dictionary mapping name -> DataFrame')

        # check if the dataset exceeds size limits
        # @todo: why is checking for filesize even an option, shouldn't it be default?
        abort_upload = self._abort_dataset_upload(
            dataset, size_check_enabled, DATASET_MAX_ROWS
        )
        if abort_upload:
            raise RuntimeError('Dataset upload aborted.')

        if info:
            # Since we started populating stats recently, some older yamls
            # dont have it. Or the user might just supply us the basic
            # schema without stats.
            # If the user provided the schema/yaml file, ask the user to
            # re-create dataset info with:
            # info = DatasetInfo.update_stats_for_existing_schema(dataset,
            # info, max_inferred_cardinality)
            # @question: shouldn't these validations be part of DatasetInfo? (high cohesion) and we invoke the method here.
            for column in info.columns:
                if (
                    (column.value_range_min is None) or (column.value_range_max is None)
                ) and column.data_type.is_numeric():
                    raise ValueError(
                        f'Dataset info does not contain min/max values for the numeric feature {column.name}. '
                        f'Please update using fdl.DatasetInfo.update_stats_for_existing_schema() '
                        f'and upload dataset with the updated dataset info.'
                    )
                if (not column.possible_values) and (
                    column.data_type.value
                    in [DataType.CATEGORY.value, DataType.BOOLEAN.value]
                ):
                    raise ValueError(
                        f'Dataset info does not contain possible values for the categorical feature {column.name}. '
                        f'Please update using fdl.DatasetInfo.update_stats_for_existing_schema() '
                        f'and upload dataset with the updated dataset info.'
                    )

            # Validate column names
            for name, df in dataset.items():
                for info_column in info.columns:
                    if info_column.name not in df:
                        raise RuntimeError(
                            f'{info_column.name}({name}) column not found in the dataframe, '
                            f'but passed in the info'
                        )

        # use inferred info with a warning if not `info` is passed
        else:
            inferred_info = DatasetInfo.from_dataframe(
                df=dataset.values(),
                display_name=dataset_id,
                dataset_id=dataset_id,
                max_inferred_cardinality=1000,
            )
            # @question: we only do validation when info is passed and not when it is inferred? Why should this be the case?
            LOG.warning(
                f'Heads up! We are inferring the details of your dataset from '
                f'the dataframe(s) provided. Please take a second to check '
                f'our work.'
                f'\n\nIf the following DatasetInfo is an incorrect '
                f'representation of your data, you can construct a '
                f'DatasetInfo with the DatasetInfo.from_dataframe() method '
                f'and modify that object to reflect the correct details of '
                f'your dataset.'
                f'\n\nAfter constructing a corrected DatasetInfo, please '
                f're-upload your dataset with that DatasetInfo object '
                f'explicitly passed via the `info` parameter of '
                f'FiddlerApi.upload_dataset().'
                f'\n\nYou may need to delete the initially uploaded version'
                f"via FiddlerApi.delete_dataset('{dataset_id}')."
                f'\n\nInferred DatasetInfo to check:'
                f'\n{textwrap.indent(repr(inferred_info), "  ")}'
            )
            info = inferred_info

        if self.strict_mode:
            info.validate()

        return self._upload_dataset_with_compression(
            project_id=project_id,
            dataset=dataset,
            dataset_id=dataset_id,
            dataset_info=info,
        )

    def upload_dataset_from_dir(
        self,
        project_id: str,
        dataset_id: str,
        dataset_dir: Path,
        file_type: str = 'csv',
        file_schema=None,
        size_check_enabled: bool = False,
    ):
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)
        dataset_dir = type_enforce('dataset_dir', dataset_dir, Path)

        if f'.{file_type}' not in SUPPORTABLE_FILE_EXTENSIONS:
            raise ValueError(
                f'Invalid file_type :{file_type}. Valid file types are : {SUPPORTABLE_FILE_EXTENSIONS}'
            )

        if file_type.endswith('avro'):
            # TODO: This was missing the last two positional arguments,
            # size_check_enabled and info; added size_check_enabled, None for
            # now.
            return self.process_avro(
                project_id,
                dataset_id,
                dataset_dir,
                file_schema,
                size_check_enabled,
                None,
            )

        if not dataset_dir.is_dir():
            raise ValueError(f'{dataset_dir} is not a directory')

        dataset_yaml = dataset_dir / f'{dataset_id}.yaml'

        if not dataset_yaml.is_file():
            raise ValueError(f'YAML file not found: {dataset_yaml}')

        with dataset_yaml.open() as f:
            dataset_info = DatasetInfo.from_dict(yaml.safe_load(f))

        files = dataset_dir.glob('*.csv')
        csv_files = [x for x in files if x.is_file()]

        LOG.info(f'Found CSV file {csv_files}')

        # Lets make sure that we add stats if they are not already there.
        # We need to read the datasets in pandas and create a dataset dictionary
        dataset = {}
        csv_paths = []
        for file in csv_files:
            csv_name = str(file).split('/')[-1]
            csv_paths.append(csv_name)
            name = csv_name[:-4]

            # @TODO Change the flow so that we can read the CSV in chunks
            dataset[name] = pd.read_csv(file, dtype=dataset_info.get_pandas_dtypes())

        # check if the dataset exceeds size limits
        abort_upload = self._abort_dataset_upload(
            dataset, size_check_enabled, DATASET_MAX_ROWS
        )
        if abort_upload:
            raise RuntimeError('Dataset upload aborted.')

        # Update stats
        dataset_info = DatasetInfo.update_stats_for_existing_schema(
            dataset, dataset_info
        )
        updated_infos = []
        for item in dataset.values():
            update_info = DatasetInfo.check_and_update_column_info(dataset_info, item)
            updated_infos.append(update_info)

        dataset_info = DatasetInfo.as_combination(
            updated_infos, display_name=dataset_info.display_name
        )

        return self._upload_dataset_with_compression(
            project_id=project_id,
            dataset=dataset,
            dataset_id=dataset_id,
            dataset_info=dataset_info,
        )

    def upload_dataset_from_file(
        self,
        project_id: str,
        dataset_id: str,
        file_path: str,
        file_type: str = 'csv',
        file_schema=Dict[str, Any],
        info: Optional[DatasetInfo] = None,
        size_check_enabled: bool = False,
    ):
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)
        file_path = type_enforce('dataset_dir', file_path, Path)

        if file_type.endswith('avro'):
            return self.process_avro(
                project_id,
                dataset_id,
                pathlib.Path(file_path),
                file_schema,
                size_check_enabled,
                info,
            )
        elif file_type.endswith('csv'):
            return self.process_csv(
                project_id,
                dataset_id,
                pathlib.Path(file_path),
                size_check_enabled,
                info,
            )

        raise ValueError(
            f'Invalid file_type :{file_type}. Valid file types are : {SUPPORTABLE_FILE_EXTENSIONS}'
        )

    def process_csv(
        self, project_id, dataset_id, csv_file_path: Path, size_check_enabled, info
    ):
        dataset = {}
        # @TODO Read and process the CSV file in chunks
        dataset[csv_file_path.name] = pd.read_csv(csv_file_path)
        return self.upload_dataset(
            project_id,
            dataset,
            dataset_id,
            size_check_enabled=size_check_enabled,
            info=info,
        )

    def process_avro(
        self,
        project_id: str,
        dataset_id: str,
        avro_file_path: Path,
        file_schema: Dict,
        size_check_enabled,
        info,
    ):
        dataset = {}
        LOG.info(f'avro_file : {avro_file_path}')
        with open(avro_file_path, 'rb') as fh:
            buf = BytesIO(fh.read())
            files = {avro_file_path: FileStorage(buf, str(avro_file_path))}
            results = upload_dataset(files, 'LOCAL_DISK', 'avro', file_schema)
            df = pd.DataFrame(results)
            dataset[avro_file_path.name] = df
            return self.upload_dataset(
                project_id=project_id,
                dataset_id=dataset_id,
                dataset=dataset,
                size_check_enabled=size_check_enabled,
                info=info,
            )

    def _upload_dataset_with_compression(
        self,
        project_id: str,
        dataset: Dict[str, pd.DataFrame],
        dataset_id: str,
        dataset_info: DatasetInfo,
    ):
        schema = dataset_info.get_arrow_schema()

        # dump the data to named parquet temp file
        with tempfile.TemporaryDirectory() as tmp:
            file_paths = []
            for name, df in dataset.items():
                # Adding .csv to support legacy way of converting everything to CSV
                filename = f'{name}.{CSV_EXTENSION}.{PARQUET_EXTENSION}'
                file_path = Path(tmp) / filename
                file_paths.append(file_path)

                # Data type conversion as per dataset_info
                # Data types like date/time/datetime should be converted to string before creating parquet file
                df = df.astype(dtype=dataset_info.get_pandas_dtypes())

                write_dataframe_to_parquet_file(
                    df=df, file_path=file_path, schema=schema
                )

            # add files to the DatasetInfo on the fly
            dataset_info = copy.deepcopy(dataset_info)
            dataset_info.files = [
                fp.name.replace(f'.{PARQUET_EXTENSION}', '') for fp in file_paths
            ]

            # upload the parquet files
            LOG.info(f'[{dataset_id}] dataset upload: upload and import dataset files')

            res = self._upload_dataset_files(
                project_id=project_id,
                dataset_id=dataset_id,
                file_paths=file_paths,
                dataset_info=dataset_info,
            )

            LOG.info(f'Dataset uploaded {res}')
            return res

    ##### End: Methods related to uploading dataset #####

    def run_model(
        self,
        project_id: str,
        model_id: str,
        df: pd.DataFrame,
        log_events=False,
        casting_type=False,
    ) -> pd.DataFrame:
        """Executes a model in the Fiddler engine on a DataFrame.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.
        :param df: A dataframe contining model inputs as rows.
        :param log_events: Variable determining if the the predictions
            generated should be logged as production traffic
        :param casting_type: Bool indicating if fiddler should try to cast the data in the event with
        the type referenced in model info. Default to False.

        :returns: A pandas DataFrame containing the outputs of the model.
        """
        return (
            self.project(project_id)
            .model(model_id)
            .predict(df, log_events, casting_type)
        )

    def run_explanation(
        self,
        project_id: str,
        model_id: str,
        df: pd.DataFrame,
        explanations: Union[str, Iterable[str]] = 'shap',
        dataset_id: Optional[str] = None,
        n_permutation: Optional[int] = None,
        n_background: Optional[int] = None,
        casting_type: Optional[bool] = False,
        return_raw_response=False,
    ) -> Union[
        AttributionExplanation,
        MulticlassAttributionExplanation,
        List[AttributionExplanation],
        List[MulticlassAttributionExplanation],
    ]:
        """Executes a model in the Fiddler engine on a DataFrame.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.
        :param df: A dataframe containing model inputs as rows. Only the first
            row will be explained.
        :param explanations: A single string or list of strings specifying
            which explanation algorithms to run.
        :param dataset_id: The unique identifier of the dataset in the
            Fiddler engine. Required for most tabular explanations, but
            optional for most text explanations.
        :param n_permutation: Number of permutations used for Fiddler SHAP. Can be used for both tabular and text data.
            By default (None), we use max(500, 2 * n_features), where n_feature is the number of word tokens
            for text input data.
        :param n_background: Number of background observations used for Fiddler SHAP for tabular data.
            By default (None), we use min(dataset.shape[0], 200)
        :param casting_type: Bool indicating if fiddler should try to cast the data in the event with
        the type referenced in model info. Default to False.

        :returns: A single AttributionExplanation if `explanations` was a
            single string, or a list of AttributionExplanation objects if
            a list of explanations was requested.
        """
        return (
            self.project(project_id)
            .model(model_id)
            .explanation(
                df,
                explanations,
                dataset_id,
                n_permutation,
                n_background,
                casting_type,
                return_raw_response,
            )
        )

    def run_feature_importance(
        self,
        project_id: str,
        model_id: str,
        dataset_id: str,
        dataset_splits: Optional[List[str]] = None,
        slice_query: Optional[str] = None,
        impact_not_importance: Optional[bool] = False,
        **kwargs,
    ) -> Dict[str, Any]:
        """Get global feature importance for a model over a dataset.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.
        :param dataset_id: The unique identifier of the dataset in the
            Fiddler engine.
        :param dataset_splits: If specified, importance will only be computed
            over these splits. Otherwise, all splits will be used. Only a
            single split is currently supported.
        :param slice_query: A special SQL query.
        :param impact_not_importance: Boolean flag to compute either impact or importance.
        False by default (compute feature importance).
        For text models, only feature impact is implemented.
        :param kwargs: Additional parameters to be passed to the importance
            algorithm. For example, `n_inputs`, `n_iterations`, `n_references`,
            `ci_confidence_level`.
        :return: A named tuple with the explanation results.
        """
        return (
            self.project(project_id)
            .model(model_id)
            .feature_importance(
                dataset_id, dataset_splits, slice_query, impact_not_importance, **kwargs
            )
        )

    def run_fairness(
        self,
        project_id: str,
        model_id: str,
        dataset_id: str,
        protected_features: list,
        positive_outcome: Union[str, int],
        slice_query: Optional[str] = None,
        score_threshold: Optional[float] = 0.5,
    ) -> Dict[str, Any]:
        """Get fairness metrics for a model over a dataset.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.
        :param dataset_id: The unique identifier of the dataset in the
            Fiddler engine.
        :param protected_features: List of protected features
        :param positive_outcome: Name or value of the positive outcome
        :param slice_query: If specified, slice the data.
        :param score_threshold: positive score threshold applied to get outcomes
        :return: A dictionary with the fairness metrics, technical_metrics,
        labels distribution and model outcomes distribution
        """
        return (
            self.project(project_id)
            .model(model_id)
            .fairness(
                dataset_id,
                protected_features,
                positive_outcome,
                slice_query,
                score_threshold,
            )
        )

    def get_mutual_information(
        self,
        project_id: str,
        dataset_id: str,
        features: list,
        normalized: Optional[bool] = False,
        slice_query: Optional[str] = None,
        sample_size: Optional[int] = None,
        seed: Optional[float] = 0.25,
    ):
        """
        The Mutual information measures the dependency between two random variables.
        It's a non-negative value. If two random variables are independent MI is equal to zero.
        Higher MI values means higher dependency.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param dataset_id: The unique identifier of the dataset in the
            Fiddler engine.
        :param features: list of features to compute mutual information with respect to all the variables in the dataset.
        :param normalized: If set to True, it will compute Normalized Mutual Information (NMI)
        :param slice_query: Optional slice query
        :param sample_size: Optional sample size for the selected dataset
        :param seed: Optional seed for sampling
        :return: a dictionary of mutual information w.r.t the given features.
        """
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)

        if isinstance(features, str):
            features = [features]
        if not isinstance(features, list):
            raise ValueError(
                f'Invalid type: {type(features)}. Argument features has to be a list'
            )
        correlation = {}
        for col_name in features:
            payload = dict(
                col_name=col_name,
                normalized=normalized,
                slice_query=slice_query,
                sample_size=sample_size,
                seed=seed,
            )
            path = ['dataset_mutual_information', self.org_id, project_id, dataset_id]
            res = self._call(path, json_payload=payload)
            correlation[col_name] = res
        return correlation

    @deprecated(
        reason='Please use add_project, this method will be removed in future versions'
    )
    def create_project(self, project_id: str):
        """Create a new project.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine. Must be a short string without whitespace.

        :returns: Server response for creation action.
        """
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)

        safe_name_check(project_id, constants.MAX_ID_LEN, self.strict_mode)
        res = None
        try:
            path = ['new_project', self.org_id, project_id]
            res = self._call(path)
        except Exception as e:
            if 'already exists' in str(e):
                LOG.error(
                    'Project name already exists, please try with a different name (You may not have access to all the projects)'
                )
            else:
                raise e

        return res

    def share_project(
        self,
        project_name: str,
        role: str,
        user_name: Optional[str] = None,
        team_name: Optional[str] = None,
    ):
        """Share a project with other users and/or teams.

        :param project_name: The name of the project to share.
        :param role: one of ["READ", "WRITE", "OWNER"].
        :param user_name: (optional) username, typically an email address.
        :param team_name: (optional) name of the team.

        :returns: Server response for creation action.
        """
        return self.project(project_name).share(role, user_name, team_name)

    def unshare_project(
        self,
        project_name: str,
        role: str,
        user_name: Optional[str] = None,
        team_name: Optional[str] = None,
    ):
        """un-Share a project with other users and/or teams.

        :param project_name: The name of the project.
        :param role: one of ["READ", "WRITE", "OWNER"].
        :param user_name: (optional) username, typically an email address.
        :param team_name: (optional) name of the team.

        :returns: Server response for creation action.
        """
        return self.project(project_name).unshare(role, user_name, team_name)

    def list_org_roles(self):
        """List the users in the organization.

        :returns: list of users and their roles in the organization.
        """
        path = ['roles', self.org_id]
        return self._call(path, is_get_request=True)

    def list_project_roles(self, project_name: str):
        """List the users and teams with access to a given project.

        :returns: list of users and teams with access to a given project.
        """
        return self.project(project_name).list_roles()

    def list_teams(self):
        """List the teams and the members in each team.

        :returns: dictionary with teams as keys and list of members as values.
        """
        path = ['teams', self.org_id]
        return self._call(path, is_get_request=True)

    ##### Start: Methods related to uploading / registering model #####

    def _import_model_predictions(
        self,
        project_id: str,
        dataset_id: str,
        model_id: str,
        columns: Sequence[Dict],
        file_paths: List[Path],
    ):
        """Uploads model predictions to Fiddler platform."""
        payload: Dict[str, Any] = dict(dataset=dataset_id)
        payload['model'] = model_id
        payload['columns'] = columns

        path = ['import_model_predictions', self.org_id, project_id]
        result = self._call(path, json_payload=payload, files=file_paths)
        return result

    def _create_model(
        self,
        project_id: str,
        dataset_id: str,
        target: str,
        features: Optional[List[str]] = None,
        train_splits: Optional[List[str]] = None,
        model_id: str = 'fiddler_generated_model',
        model_info: Optional[ModelInfo] = None,
    ):
        """Trigger auto-modeling on a dataset already uploaded to Fiddler.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.

        :param dataset_id: The unique identifier of the dataset in the
            Fiddler engine.
        :param target: The column name of the target to be modeled.
        :param features: If specified, a list of column names to use as
            features. If not specified, all non-target columns will be used
            as features.
        :param train_splits: A list of splits to train on. If not specified,
            all splits will be used as training data. Currently only a single
            split can be specified if `train_splits` is not None.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.

        :returns: Server response for creation action.
        """
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)

        safe_name_check(model_id, constants.MAX_ID_LEN, self.strict_mode)

        if train_splits is not None and len(train_splits) > 1:
            raise NotImplementedError(
                'Sorry, currently only single-split training is '
                'supported. Please only pass a maximum of one element to '
                '`train_splits`.'
            )
        source = None if train_splits is None else train_splits[0]
        dataset_column_names = self.get_dataset_info(
            project_id, dataset_id
        ).get_column_names()

        # raise exception if misspelled target
        if target not in dataset_column_names:
            raise ValueError(
                f'Target "{target}" not found in the columns of '
                f'dataset {dataset_id} ({dataset_column_names})'
            )

        # raise if target in features or features not in columns
        if features is not None:
            if target in features:
                raise ValueError(f'Target "{target}" cannot also be in ' f'features.')
            features_not_in_dataset = set(features) - set(dataset_column_names)
            if len(features_not_in_dataset) > 0:
                raise ValueError(
                    f'All features should be in the dataset, but '
                    f'the following features were not found in '
                    f'the dataset: {features_not_in_dataset}'
                )

        # use all non-target columns from dataset if no features are specified
        if features is None:
            features = list(dataset_column_names)
            features.remove(target)

        payload: Dict[str, Any] = {
            'dataset': dataset_id,
            'model_name': model_id,
            'source': source,
            'inputs': features,
            'targets': [target],
        }

        if model_info:
            if self.strict_mode:
                model_info.validate()
            payload['model_info'] = dict(model=model_info.to_dict())

        path = ['generate', self.org_id, project_id]
        result = self._call_executor_service(path, json_payload=payload)
        return result

    def _upload_model_custom(
        self,
        artifact_path: Path,
        model_info: ModelInfo,
        project_id: str,
        model_id: str,
        associated_dataset_ids: Optional[List[str]] = None,
        deployment_options: Optional[DeploymentOptions] = None,
    ):
        """Uploads a custom model object to the Fiddler engine along with
            custom glue-code for running the model. Optionally, a new runtime
            (k8s deployment) can be specified for the model via
            the deployment_options.

            Note: The parameters namespace, port, replicas, cpus, memory, gpus,
            await_deployment are only used if an image_uri is specified.

        :param artifact_path: A path to a directory containing all of the
            model artifacts needed to run the model. This includes a
            `package.py` file with the glue code needed to run the model.
        :param model_info: A ModelInfo object describing the details of the model.
        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine. Must be a short string without
            whitespace.
        :param associated_dataset_ids: The unique identifiers of datasets in
            the Fiddler engine to associate with the model.
        :param deployment_options: Options to control various deployment characteristics like deployment_type,
            whether to block until deployment completes etc.

        :returns: Server response for upload action.
        """
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)

        safe_name_check(model_id, constants.MAX_ID_LEN, self.strict_mode)

        if not artifact_path.is_dir():
            raise ValueError(f'The {artifact_path} must be a directory.')

        model_info = FiddlerApi._add_dataset_ids_to_model_info(
            model_info, associated_dataset_ids
        )

        if self.strict_mode:
            model_info.validate()

        if not deployment_options:
            deployment_options = DeploymentOptions(
                deployment_type=DeploymentType.PREDICTOR
            )

        # upload the model
        payload = dict(
            project=project_id,
            model=model_id,
            model_schema=dict(model=model_info.to_dict()),
            framework=model_info.framework,
            upload_as_archive=True,
            model_type='custom',
        )

        # Add deployment_options
        payload.update(deployment_options.to_dict())

        with tempfile.TemporaryDirectory() as tmp:
            tarfile_path = Path(tmp) / 'files'
            shutil.make_archive(
                base_name=str(Path(tmp) / 'files'),
                format='tar',
                root_dir=str(artifact_path),
                base_dir='.',
            )
            LOG.info(
                f'[{model_id}] model upload: uploading custom model from'
                f' artifacts in {str(artifact_path)} tarred to '
                f'{str(tarfile_path)}'
            )

            endpoint_path = ['model_upload', self.org_id, project_id]
            result = self._call(
                endpoint_path, json_payload=payload, files=[Path(tmp) / 'files.tar']
            )
            return result

    def upload_model_package(
        self,
        artifact_path: Path,
        project_id: str,
        model_id: str,
        deployment_type: Optional[
            str
        ] = DeploymentType.PREDICTOR,  # model deployment type. One of {'predictor', 'executor'}
        image_uri: Optional[str] = None,  # image to be used for newly uploaded model
        namespace: Optional[str] = None,  # kubernetes namespace
        port: Optional[int] = 5100,  # port on which model is served
        replicas: Optional[int] = 1,  # number of replicas
        cpus: Optional[float] = 0.25,  # number of CPU cores
        memory: Optional[str] = '128m',  # amount of memory required.
        gpus: Optional[int] = 0,  # number of GPU cores
        await_deployment: Optional[bool] = True,  # wait for deployment
    ):
        """Uploads a custom model package to the Fiddler engine along with
            custom glue-code for running the model. Optionally, a new runtime
            (k8s deployment) can be specified for the model via
            the deployment_type and the image_uri parameters.

            Note: The parameters namespace, port, replicas, cpus, memory, gpus,
            await_deployment are only used if an image_uri is specified.

        :param artifact_path: A path to a directory containing all of the
            model artifacts needed to run the model. This includes a
            `package.py` file with the glue code needed to run the model.
        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine. Must be a short string without
            whitespace.
        :param deployment_type: One of {'predictor', 'executor'}
        'predictor': where the model just exposes a `/predict` endpoint
                     - typically simple sklearn like models
        'executor': where fiddler needs the model internals
                     - typically deep models like tensorflow and pytorch etc
        :param image_uri: A URI of the form <registry>/<image-name>:<tag> which
            if specified will be used to create a new runtime and then serve the
            model.
        :param namespace: The kubernetes namespace to use for the newly created
            runtime.
        :param port: The port to use for the newly created runtime.
        :param replicas: The number of replicas running the model.
        :param cpus: The number of CPU cores reserved per replica.
        :param memory: The amount of memory reservation per replica.
        :param gpus: The number of GPU cores reserved per replica.
        :param await_deployment: whether to block until deployment completes.

        :returns: Server response for upload action.
        """

        # Type enforcement
        artifact_path = type_enforce('artifact_path', artifact_path, Path)

        if not artifact_path.is_dir():
            raise ValueError(f'Not a valid model dir: {artifact_path}')

        yaml_file = artifact_path / 'model.yaml'
        if not yaml_file.is_file():
            raise ValueError(f'Model yaml not found {yaml_file}')

        with yaml_file.open() as f:
            model_info = ModelInfo.from_dict(yaml.safe_load(f))

        deployment_type = DeploymentType[deployment_type.upper()]
        valid_deployment_types = [DeploymentType.PREDICTOR, DeploymentType.EXECUTOR]
        if deployment_type not in valid_deployment_types:
            raise ValueError(f'Pass valid deployment type({valid_deployment_types})')

        deployment_options = DeploymentOptions(
            deployment_type=deployment_type,
            image=image_uri,
            namespace=namespace,
            port=port,
            replicas=replicas,
            cpus=cpus,
            memory=memory,
            gpus=gpus,
            await_deployment=await_deployment,
        )

        miv = ModelInfoValidator(model_info, None)
        miv.validate_categoricals(modify=True)
        self._upload_model_custom(
            artifact_path=artifact_path,
            project_id=project_id,
            model_id=model_id,
            model_info=model_info,
            deployment_options=deployment_options,
        )

    @staticmethod
    def _add_dataset_ids_to_model_info(model_info, associated_dataset_ids):
        model_info = copy.deepcopy(model_info)
        # add associated dataset ids to ModelInfo
        if associated_dataset_ids is not None:
            for dataset_id in associated_dataset_ids:
                assert (
                    ' ' not in dataset_id
                ), 'Dataset identifiers should not contain whitespace'
            model_info.misc['datasets'] = associated_dataset_ids
        return model_info

    def _trigger_model_predictions(
        self, project_id: str, model_id: str, dataset_id: str
    ):
        """Makes the Fiddler service compute and cache model predictions on a
        dataset."""
        return (
            self.project(project_id)
            .model(model_id)
            ._trigger_model_predictions(dataset_id)
        )

    def trigger_pre_computation(
        self,
        project_id: str,
        model_id: str,
        dataset_id: str,
        overwrite_cache: Optional[bool] = False,
        batch_size: Optional[int] = 10,
        calculate_predictions: Optional[bool] = True,
        cache_global_pdps: Optional[bool] = False,
        cache_global_impact_importance: Optional[bool] = True,
        cache_dataset=False,
    ):
        """Triggers various precomputation steps within the Fiddler service based on input parameters.

        :param project_id:                        the project to which the model whose events are
                                                  being published belongs.
        :param model_id:                          the model whose events are being published.
        :param dataset_id:                        id of the dataset to be used.
        :param overwrite_cache:                   Boolean indicating whether to overwrite previously cached
                                                  information.
        :param batch_size:                        Batch size of global PDP calculation.
        :param calculate_predictions:             Boolean indicating whether to pre-calculate and store model
                                                  predictions.
        :param cache_global_pdps:                 Boolean indicating whether to pre-calculate and cache global partial
                                                  dependence plots.
        :param cache_global_impact_importance:    Boolean indicating whether to pre-calculate and global feature impact
                                                  and global feature importance.
        :param cache_dataset:                     Boolean indicating whether to cache dataset histograms.
                                                  Should be set to True for large datasets.
        """
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)

        path = ['trigger_pre_computation', self.org_id, project_id, model_id]
        payload = {
            'dataset_id': dataset_id,
            'calculate_predictions': calculate_predictions,
            'cache_global_pdps': cache_global_pdps,
            'cache_global_impact_importance': cache_global_impact_importance,
            'overwrite_cache': overwrite_cache,
            'batch_size': batch_size,
            'cache_dataset': cache_dataset,
        }
        try:
            result = self._call_executor_service(path, payload, stream=True)
            for res in result:
                print_streamed_result(res)
        except Exception as e:
            LOG.exception('Failed while trigger precomputation, error message: ')
            raise e

    def _deploy_container_model(
        self,
        project_id: str,
        model_id: str,
        dataset_id: str,
        model_info: ModelInfo,
        deployment_options: DeploymentOptions,
    ):
        if deployment_options.deployment_type == DeploymentType.FAR:
            deployment_options.deployment_type = DeploymentType.EXECUTOR
        else:
            deployment_options.deployment_type = DeploymentType.PREDICTOR
            if model_info.mlflow_params is None:
                port = 8080
                if deployment_options.port:
                    port = deployment_options.port
                host_name = model_string_to_hostname(
                    f'{self.org_id}-{project_id}-{model_id}'
                )
                default_mlflow_url = f'http://{host_name}:{port}/invocations'
                model_info.mlflow_params = MLFlowParams(
                    relative_path_to_saved_model='.', live_endpoint=default_mlflow_url
                )

        package_py = (
            'def get_model():\n  '
            'raise ValueError("This should not called. '
            'Use mlflow model container instead")'
        )

        with tempfile.TemporaryDirectory() as tmp:
            artifact_path = Path(tmp)
            package_file = artifact_path / 'package.py'
            package_file.write_text(package_py)

            return self._upload_model_custom(
                artifact_path=artifact_path,
                project_id=project_id,
                model_id=model_id,
                model_info=model_info,
                associated_dataset_ids=[dataset_id],
                deployment_options=deployment_options,
            )

    def register_model(
        self,
        project_id: str,
        model_id: str,
        dataset_id: str,
        model_info: ModelInfo,
        deployment: Optional[DeploymentOptions] = None,
        cache_global_impact_importance: bool = True,
        cache_global_pdps: bool = False,
        cache_dataset: bool = True,
    ):
        """
        Register a model in fiddler. This will generate a surrogate model,
        which can be replaced later with original model.

        Note: This method can take a while if the dataset is large. It is
        recommended to call register_model on a smaller representative
        dataset, before trying out on larger dataset.

        :param project_id: id of the project
        :param model_id: name to be used for the dataset and model
        :param dataset_id: id of the dataset to be used
        :param model_info: model info
        :param deployment: Model deployment options
        :param cache_global_impact_importance: Boolean indicating whether to pre-calculate and global feature impact
        and global feature importance.
        :param cache_global_pdps: Boolean indicating whether to pre-calculate and cache global partial dependence plots.
        :param cache_dataset: Boolean indicating whether to cache dataset histograms.
        Should be set to True for large datasets.
        """

        # Input Validation
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)

        safe_name_check(model_id, constants.MAX_ID_LEN, self.strict_mode)

        if self.strict_mode:
            model_info.validate()

        if not deployment:
            deployment = DeploymentOptions(deployment_type=DeploymentType.SURROGATE)

        # Register a surrogate model or Upload a far container model
        path = ['register_model', self.org_id, project_id, model_id]
        payload = {
            'dataset_id': dataset_id,
            'model_info': dict(model=model_info.to_dict()),
            'deployment_options': deployment.to_dict(),
            'cache_global_pdps': cache_global_pdps,
            'cache_global_impact_importance': cache_global_impact_importance,
            'cache_dataset': cache_dataset,
        }

        try:
            result = self._call_executor_service(path, payload, stream=True)
            for res in result:
                print_streamed_result(res)

        except Exception as e:
            LOG.exception('Failed to register model')
            raise e

    def update_model(
        self,
        project_id: str,
        model_id: str,
        model_dir: Path,
        force_pre_compute: bool = True,
    ):
        """
        update the specified model, with model binary and package.py from
        the specified model_dir

        Note: changes to model.yaml is not supported right now.

        :param project_id: project id
        :param model_id: model id
        :param model_dir: model directory
        :param force_pre_compute: if true refresh the pre-computated values.
               This can also be done manually by calling trigger_pre_computation
        :return: model artifacts
        """
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        model_dir = type_enforce('model_dir', model_dir, Path)

        if not model_dir.is_dir():
            raise ValueError(f'not a valid model directory: {model_dir}')
        yaml_file = model_dir / 'model.yaml'
        if not yaml_file.is_file():
            raise ValueError(f'Model yaml not found {yaml_file}')
        with yaml_file.open() as f:
            model_info = ModelInfo.from_dict(yaml.safe_load(f))

        if len(model_info.datasets) < 1:
            raise ValueError('Unable to find dataset in model.yaml')

        if len(model_info.datasets) > 1:
            raise ValueError('More than one dataset specified in model.yaml')
        dataset_id = model_info.datasets[0]

        remote_model_info = self.get_model_info(project_id, model_id)
        # Model info should remain exactly the same, expect the following fields can be changed:
        # custom_explanation_names, preferred_explanation_method, display_name, description, framework, algorithm
        # and model_deployment_params
        ddiff = DeepDiff(remote_model_info, model_info, ignore_order=True,
                         exclude_paths=['root.preferred_explanation_method', 'root.custom_explanation_names',
                                        'root.display_name', 'root.description', 'root.framework', 'root.algorithm',
                                        'root.model_deployment_params'])

        if len(ddiff) > 0:
            raise ValueError(
                f'remote model info, does not match '
                f'local model info: {ddiff}. Updating those fields in model info is '
                f'not currently supported'
            )

        LOG.info('Loading dataset info')
        dataset_info = self.get_dataset_info(project_id, dataset_id)

        if self.strict_mode:
            LOG.info('Validating ...')
            # todo: enable this validator. It need sklearn etc framework libs
            # in client path. So, not sure if we should do this by default
            # validator = PackageValidator(model_info, dataset_info, model_dir)
            # passed, errors = validator.run_chain()
            # if not passed:
            #     raise ValueError(f'validation failed with errors: {errors}')

            ModelInfoValidator(model_info, dataset_info).validate()
            model_info.validate()
        else:
            LOG.info('Validation skipped')

        tmp = tempfile.mkdtemp()
        shutil.make_archive(
            base_name=str(Path(tmp) / 'model_package'),
            format='tar',
            root_dir=str(model_dir),
            base_dir='.',
        )

        payload: Dict[str, Any] = {}
        endpoint_path = ['update_model', self.org_id, project_id, model_id]
        self._call(
            endpoint_path,
            json_payload=payload,
            files=[Path(tmp) / 'model_package.tar'],
        )
        LOG.info('Model updated')

        LOG.info('Testing updated model')
        should_log = self.connection.capture_server_log
        status = True
        for i in range(3):
            try:
                self.connection.capture_server_log = True
                sample_df = self._get_dataset_sample(project_id, dataset_id, 10)
                self.run_model(project_id, model_id, sample_df, log_events=False)
                LOG.info('Server Logs: ')
                LOG.info(self.connection.last_server_log)
                LOG.info('All tests passed ..')
                break
            except Exception as e:
                status = False
                if i == 2:
                    LOG.warning(f'Test failed with error: {e}')
                else:
                    LOG.warning(f'Retrying test {i}')
            finally:
                self.connection.capture_server_log = should_log

        if status and force_pre_compute:
            try:
                self.trigger_pre_computation(
                    project_id,
                    model_id,
                    dataset_id,
                    overwrite_cache=True,
                    calculate_predictions=True,
                    cache_global_pdps=True,
                    cache_global_impact_importance=True,
                )
            except Exception as e:
                LOG.warning(
                    'Model was updated successfully, but failed to refresh '
                    'pre-computed values. You can retry this operation by '
                    f'calling trigger_pre_computation(), error: {e}'
                )
                status = False

        shutil.rmtree(tmp)
        return status

    ##### End: Methods related to uploading / registering model #####

    ##### Start: Methods related to monitoring config #####

    def initialize_monitoring(  # noqa
        self,
        project_id: str,
        model_id: str,
        enable_modify: Optional[bool] = False,
        verbose: Optional[bool] = False,
        is_sync: bool = False,
        version: str = 'fiddler1',
    ):
        """
        Ensure that monitoring has been setup and Fiddler is ready to ingest events.

        :param project_id:          The project for which to initialize monitoring.
        :param model_id:            The model for which to initialize monitoring.
        :param model_id:            The model for which to initialize monitoring.
        :param enable_modify:       [DEPRECATED - only fiddler1] Grant the Fiddler backend permission to
                                    modify model related objects, schema, etc.
                                    Can be bool  `True`/`False`, indicating global
                                    write/read-only permission.
                                    Alternatively, can be a sequence of elements
                                    from `fiddler.core_objects.InitMonitoringModifications`,
                                    in which case all listed elements will be
                                    assumed to be modifiable, and omitted ones

                                    will be read-only.
        :param verbose:             [DEPRECATED - only fiddler1] Bool indicating whether to run in verbose
                                    mode with longer debug messages for errors.
        :param is_sync:             [NEW - only fiddler2] Bool indicating if monitoring must be initialized
                                    synchronously or asynchronously
        :param version:             String (either 'fiddler2' or 'fiddler1') indicating whether to use legacy functionality (with enable_modify and verbose) or new functionality (is_sync)

        :returns bool indicating whether monitoring could be setup correctly.
        """
        return 'Initialize monitoring is only supported from v2'

        # @todo: the below method needs to be called if we want init monitoring for v1 client and that needs server side fixing.

        # resp = Monitoring(self.connection, project_id, model_id).initialize_monitoring(
        #     enable_modify, verbose, version
        # )

    def add_monitoring_config(
        self,
        config_info: dict,
        project_id: Optional[str] = None,
        model_id: Optional[str] = None,
    ):
        """Adds a config for either an entire org, or project or a model.
        Here's a sample config:
        {
            'min_bin_value': 3600, # possible values 300, 3600, 7200, 43200, 86400, 604800 secs
            'time_ranges': ['Day', 'Week', 'Month', 'Quarter', 'Year'],
            'default_time_range': 7200,
            'tag': 'anything you want',
            ‘aggregation_config’: {
               ‘baseline’: {
                  ‘type’: ‘dataset’,
                  ‘dataset_name’: yyy
               }
            }
        }
        """
        return Monitoring(self.connection, project_id, model_id).add_monitoring_config(
            config_info
        )

    ##### Start: Methods related to publishing event #####

    def _basic_drift_checks(self, project_id, model_info, model_id):
        # Lets make sure prediction table is created and has prediction data by
        # just running the slice query
        violations = []
        try:
            query_str = f'select * from "{model_info.datasets[0]}.{model_id}" limit 1'
            df = self.get_slice(
                query_str,
                project_id=project_id,
            )
            for index, row in df.iterrows():
                for out_col in model_info.outputs:
                    if out_col.name not in row:
                        msg = f'Drift error: {out_col.name} not in predictions table. Please delete and re-register your model.'
                        violations.append(
                            MonitoringViolation(MonitoringViolationType.WARNING, msg)
                        )
        except RuntimeError:
            msg = 'Drift error: Predictions table does not exists. Please run trigger_pre_computation for an existing model, or use register_model to register a new model.'
            violations.append(MonitoringViolation(MonitoringViolationType.WARNING, msg))
            return violations

        return violations

    def publish_event(
        self,
        project_id: str,
        model_id: str,
        event: dict,
        event_id: Optional[str] = None,
        update_event: Optional[bool] = None,
        event_timestamp: Optional[int] = None,
        timestamp_format: FiddlerTimestamp = FiddlerTimestamp.INFER,
        casting_type: Optional[bool] = False,
        dry_run: Optional[bool] = False,
    ):
        """
        Publishes an event to Fiddler Service.
        :param project_id: The project to which the model whose events are being published belongs
        :param model_id: The model whose events are being published
        :param dict event: Dictionary of event details, such as features and predictions.
        :param event_id: Unique str event id for the event
        :param update_event: Bool indicating if the event is an update to a previously published row
        :param event_timestamp: The UTC timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        :param timestamp_format:   Format of timestamp within batch object. Can be one of:
                                - FiddlerTimestamp.INFER
                                - FiddlerTimestamp.EPOCH_MILLISECONDS
                                - FiddlerTimestamp.EPOCH_SECONDS
                                - FiddlerTimestamp.ISO_8601
        :param casting_type: Bool indicating if fiddler should try to cast the data in the event with
        the type referenced in model info. Default to False.
        :param dry_run: If true, the event isnt published and instead the user gets a report which shows
        IF the event along with the model would face any problems with respect to monitoring

        """
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        event = type_enforce('event', event, dict)
        if event_id:
            event_id = type_enforce('event_id', event_id, str)

        if casting_type:
            try:
                model_info = self.get_model_info(project_id, model_id)
            except RuntimeError:
                raise RuntimeError(
                    f'Did not find ModelInfo for project "{project_id}" and model "{model_id}".'
                )
            event = cast_input_data(event, model_info)

        if not timestamp_format or timestamp_format not in FiddlerTimestamp:
            raise ValueError('Please specify a valid timestamp_format')

        assert timestamp_format is not None, 'timestamp_format unexpectedly None'
        event['__timestamp_format'] = timestamp_format.value

        if update_event:
            event['__event_type'] = 'update_event'
            event['__updated_at'] = event_timestamp
            if event_id is None:
                raise ValueError('An update event needs an event_id')
        else:
            event['__event_type'] = 'execution_event'
            event['__occurred_at'] = event_timestamp

        if event_id is not None:
            event['__event_id'] = event_id

        if dry_run:
            violations = self._pre_flight_monitoring_check(project_id, model_id, event)
            violations_list = []
            LOG.info('\n****** publish_event dry_run report *****')
            LOG.info(f'Found {len(violations)} Violations:')
            for violation in violations:
                violations_list.append(
                    {'type': violation.type.value, 'desc': violation.desc}
                )
                LOG.info(f'Type: {violation.type.value: <11}{violation.desc}')
            result = json.dumps(violations_list)
        else:
            path = ['external_event', self.org_id, project_id, model_id]
            # The ._call uses `timeout` and `num_tries` logic due to an issue with Mac/Docker.
            # This is only enabled using the env variable `FIDDLER_RETRY_PUBLISH`; otherwise it
            # is a normal ._call function
            result = self._call(path, event, timeout=2, num_tries=5)

        return result

    def _pre_flight_monitoring_check(self, project_id, model_id, event):
        violations = []
        violations += self._basic_monitoring_tests(project_id, model_id)
        if len(violations) == 0:
            model_info = self.get_model_info(project_id, model_id)
            dataset_info = self.get_dataset_info(project_id, model_info.datasets[0])
            violations += self._basic_drift_checks(project_id, model_info, model_id)
            violations += self.monitoring_validator.pre_flight_monitoring_check(
                event, model_info, dataset_info
            )
        return violations

    def _basic_monitoring_tests(self, project_id, model_id):
        """Basic checks which would prevent monitoring from working altogether."""
        violations = []
        try:
            model_info = self.get_model_info(project_id, model_id)
        except RuntimeError:
            msg = f'Error: Model:{model_id} in project:{project_id} does not exist'
            violations.append(MonitoringViolation(MonitoringViolationType.FATAL, msg))
            return violations

        try:
            _ = self.get_dataset_info(project_id, model_info.datasets[0])
        except RuntimeError:
            msg = f'Error: Dataset:{model_info.datasets[0]} does not exist'
            violations.append(MonitoringViolation(MonitoringViolationType.FATAL, msg))
            return violations

        return violations

    def publish_events_log(
        self,
        project_id: str,
        model_id: str,
        logs: pd.DataFrame,
        force_publish: Optional[bool] = None,
        ts_column: Optional[str] = None,
        default_ts: Optional[int] = None,
        num_threads: int = 1,
        batch_size: Optional[int] = None,
        casting_type: Optional[bool] = False,
    ):
        """
        [SOON TO BE DEPRECATED] Publishes prediction log to Fiddler Service.
        :param project_id:    The project to which the model whose events are being published belongs
        :param model_id:      The model whose events are being published
        :param logs:          Data frame of event details, such as features and predictions.
        :param force_publish: Continue with publish even if all input and output columns not in log.
        :param ts_column:     Column to extract timestamp value from.
                              Timestamp must be UTC in format: `%Y-%m-%d %H:%M:%S.%f` (e.g. 2020-01-01 00:00:00.000000)
                              or as the timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        :param default_ts:    Default timestamp to use if ts_column not specified.
                              Must be given as the timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        :param num_threads:   Number of threads to parallelize this function over. Dataset will be divided evenly
                              over the number of threads.
        :param batch_size:    Size of dataframe to publish in any given run.
        :param casting_type: Bool indicating if fiddler should try to cast the data in the event with
        the type referenced in model info. Default to False.

        """
        return PublishEvent.publish_events_log(
            project_id,
            model_id,
            logs,
            force_publish,
            ts_column,
            default_ts,
            num_threads,
            batch_size,
            casting_type,
        )

    def publish_events_batch(  # noqa
        self,
        project_id: str,
        model_id: str,
        batch_source: Union[pd.DataFrame, str],
        id_field: Optional[str] = None,
        update_event: Optional[bool] = False,
        timestamp_field: Optional[str] = None,
        timestamp_format: FiddlerTimestamp = FiddlerTimestamp.INFER,
        data_source: Optional[BatchPublishType] = None,
        casting_type: Optional[bool] = False,
        credentials: Optional[dict] = None,
        group_by: Optional[str] = None,
    ):
        """
        Publishes a batch events object to Fiddler Service.
        :param project_id:    The project to which the model whose events are being published belongs.
        :param model_id:      The model whose events are being published.
        :param batch_source:  Batch object to be published. Can be one of: Pandas DataFrame, CSV file, PKL Pandas DataFrame, or Parquet file.
        :param id_field:  Column to extract id value from.
        :param update_event: Bool indicating if the events are updates to previously published rows
        :param timestamp_field:     Column to extract timestamp value from.
                              Timestamp must match the specified format in `timestamp_format`.
        :param timestamp_format:   Format of timestamp within batch object. Can be one of:
                                - FiddlerTimestamp.INFER
                                - FiddlerTimestamp.EPOCH_MILLISECONDS
                                - FiddlerTimestamp.EPOCH_SECONDS
                                - FiddlerTimestamp.ISO_8601
        :param data_source:   Source of batch object. In case of failed inference, can be one of:
                                - BatchPublishType.DATAFRAME
                                - BatchPublishType.LOCAL_DISK
                                - BatchPublishType.AWS_S3
                                - BatchPublishType.GCP_STORAGE
        :param casting_type: Bool indicating if fiddler should try to cast the data in the event with
                             the type referenced in model info. Default to False.
        :param credentials:  Dictionary containing authorization for AWS or GCP.

                             For AWS S3, list of expected keys are
                              ['aws_access_key_id', 'aws_secret_access_key', 'aws_session_token']
                              with 'aws_session_token' being applicable to the AWS account being used.

                             For GCP, list of expected keys are
                              ['gcs_access_key_id', 'gcs_secret_access_key', 'gcs_session_token']
                              with 'gcs_session_token' being applicable to the GCP account being used.
        :param group_by: Column to group events together for Model Performance metrics. For example,
                         in ranking models that column should be query_id or session_id, used to
                         compute NDCG and MAP. Be aware that the batch_source file/dataset provided should have
                         events belonging to the SAME query_id/session_id TOGETHER and cannot be mixed
                         in the file. For example, having a file with rows belonging to query_id 31,31,31,2,2,31,31,31
                         would not work. Please sort the file by group_by group first to have rows with
                         the following order: query_id 31,31,31,31,31,31,2,2.
        """
        return PublishEvent(self.connection).publish_events_batch(
            project_id,
            model_id,
            batch_source,
            id_field,
            update_event,
            timestamp_field,
            timestamp_format,
            data_source,
            casting_type,
            credentials,
            group_by,
        )

    def publish_events_batch_schema(  # noqa
        self,
        batch_source: Union[pd.DataFrame, str],
        publish_schema: Dict[str, Any],
        data_source: Optional[BatchPublishType] = None,
        credentials: Optional[dict] = None,
        group_by: Optional[str] = None,
    ):
        """
        Publishes a batch events object to Fiddler Service.
        :param batch_source:  Batch object to be published. Can be one of: Pandas DataFrame, CSV file, PKL Pandas DataFrame, or Parquet file.
        :param publish_schema: Dict object specifying layout of data.
        :param data_source:   Source of batch object. In case of failed inference, can be one of:
                                - BatchPublishType.DATAFRAME
                                - BatchPublishType.LOCAL_DISK
                                - BatchPublishType.AWS_S3
                                - BatchPublishType.GCP_STORAGE
        :param credentials:  Dictionary containing authorization for AWS or GCP.

                             For AWS S3, list of expected keys are
                              ['aws_access_key_id', 'aws_secret_access_key', 'aws_session_token']
                              with 'aws_session_token' being applicable to the AWS account being used.

                             For GCP, list of expected keys are
                              ['gcs_access_key_id', 'gcs_secret_access_key', 'gcs_session_token']
                              with 'gcs_session_token' being applicable to the GCP account being used.
        :param group_by: Column to group events together for Model Performance metrics. For example,
                         in ranking models that column should be query_id or session_id, used to
                         compute NDCG and MAP.
        """
        return PublishEvent(self.connection).publish_events_batch_schema(
            batch_source,
            publish_schema,
            data_source,
            credentials,
            group_by,
        )

    def publish_parquet_s3(
        self,
        project_id: str,
        model_id: str,
        parquet_file: str,
        auth_context: Optional[Dict[str, Any]] = None,
        ts_column: Optional[str] = None,
        default_ts: Optional[int] = None,
    ):
        """
        [SOON TO BE DEPRECATED] Publishes parquet events file from S3 to Fiddler instance. Experimental and may be expanded in the future.

        :param project_id:    The project to which the model whose events are being published belongs
        :param model_id:      The model whose events are being published
        :param parquet_file:  s3_uri for parquet file to be published
        :param auth_context:  Dictionary containing authorization for AWS. List of expected keys are
                              ['aws_access_key_id', 'aws_secret_access_key', 'aws_session_token']
                              with 'aws_session_token' being applicable to the AWS account being used.
        :param ts_column:     Column to extract time stamp value from.
                              Timestamp must be UTC in format: `%Y-%m-%d %H:%M:%S.%f` (e.g. 2020-01-01 00:00:00.000000)
                              or as the timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        :param default_ts:    Default timestamp to use if ts_column not specified.
                              Must be given as the timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        """
        return PublishEvent(self.connection).publish_parquet_s3(
            project_id, model_id, parquet_file, auth_context, ts_column, default_ts
        )

    ##### End: Methods related to publishing event #####

    def generate_sample_events(
        self,
        project_id: str,
        model_id: str,
        dataset_id: str,
        number_of_events: int = 100,
        time_range: int = 8,
    ):
        """
        Generate monitoring traffic for the given model. Traffic is generated
        by randomly sampling rows from the specified dataset.

        Note: This method can be used to generate monitoring traffic for
        testing purpose. In production, use publish_event or publish_events_logs
        to send model input and output to fiddler.

        :param project_id:
        :param model_id:
        :param dataset_id:
        :param number_of_events: number of prediction events to generate
        :param time_range: number of days. time_range is used
                to spread the traffic
        :return: sample events that can be published to fiddler
        """
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)

        if number_of_events < 1 or number_of_events > 1000:
            raise ValueError('number_of_events must be between 1 and 1000')

        if time_range < 1 or time_range > 365:
            raise ValueError('time_range must be between 1 and 365 days')

        model_info = self.get_model_info(project_id, model_id)
        if model_info.model_task.value == ModelTask.RANKING.value:
            is_ranking = True
        else:
            is_ranking = False
        event_sample_df = self._get_dataset_sample(
            project_id, dataset_id, number_of_events, is_ranking
        )

        outputs_available = all(
            c in event_sample_df for c in model_info.get_output_names()
        )

        if outputs_available:
            result_df = event_sample_df
        else:
            # get prediction result
            result = self.run_model(
                project_id, model_id, event_sample_df, log_events=False
            )
            result_df = pd.concat([event_sample_df, result], axis=1)

        # create well distributed time stamps
        ONE_DAY_MS = 8.64e7
        event_time = round(time.time() * 1000) - (ONE_DAY_MS * time_range)
        time_stamp = []

        if is_ranking:
            group_by = model_info.group_by
            result_df = result_df.sort_values(by=[group_by])
            query_num, query_count = np.unique(result_df[group_by], return_counts=True)
            interval = round((time.time() * 1000 - event_time) / len(query_num))
            for _, c in enumerate(query_count):
                time_stamp.extend(c * [event_time])
                event_time = event_time + random.randint(1, interval * 2)
        else:
            interval = round((time.time() * 1000 - event_time) / number_of_events)
            for i in range(0, len(result_df)):
                time_stamp.append(event_time)
                event_time = event_time + random.randint(1, interval * 2)
        result_df['__occurred_at'] = time_stamp

        return result_df

    def _get_dataset_sample(
        self, project_id, dataset_id, sample_size, is_ranking=False
    ):
        if is_ranking:
            sampling = False
            LOG.info(
                f'No sampling when retrieving data {dataset_id} because this is a ranking model.'
            )
        else:
            sampling = True
        dataset_dict = self.get_dataset(
            project_id, dataset_id, max_rows=sample_size, sampling=sampling
        )
        datasets = dataset_dict.values()
        for df in datasets:
            LOG.info(f'In get dataset sample the df shape: {df.shape}')
            df.reset_index(inplace=True, drop=True)
        df = pd.concat(datasets, ignore_index=True)
        if len(df) > sample_size:
            df = df[:sample_size]
        # note: len(df) can be less than sample_size
        LOG.info(f'In get dataset sample the value being returned: {df.shape}')
        return df

    def get_model(self, project_id: str, model_id: str, output_dir: Path):
        """
        download the model binary, package.py and model.yaml to the given
        output dir.

        :param project_id: project id
        :param model_id: model id
        :param output_dir: output directory
        :return: model artifacts
        """
        return self.project(project_id).model(model_id).download(output_dir)

    # Methods related to segment

    def list_segments(
        self,
        project_id: str,
        model_id: str,
        activated: Optional[bool] = None,
        schema_version: Optional[float] = None,
    ):
        """
        List all segments currently associated with a model. Can filter by activated / schema_version if desired.

        activated = [None/True/False], for all, activated-only, deactivate-only, respectively
        schema_verios = [None/0.1] for all, 0.1 respectively
        """
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        return Segment(self.connection, project_id, model_id).list_segments(
            activated, schema_version
        )

    def upload_segment(
        self,
        project_id: str,
        model_id: str,
        segment_id: str,
        segment_info: SegmentInfo,
    ):
        """Validates and uploads an existing SegmentInfo object to Fiddler."""
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        segment_id = type_enforce('segment_id', segment_id, str)
        safe_name_check(segment_id, constants.MAX_ID_LEN, self.strict_mode)

        return Segment(self.connection, project_id, model_id).upload_segment(
            segment_id, segment_info
        )

    def activate_segment(self, project_id: str, model_id: str, segment_id: str):
        """Activate an existing segment in Fiddler"""
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        segment_id = type_enforce('segment_id', segment_id, str)

        return Segment(self.connection, project_id, model_id).activate_segment(
            segment_id
        )

    def deactivate_segment(self, project_id: str, model_id: str, segment_id: str):
        """Deactivate an existing segment in Fiddler"""
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        segment_id = type_enforce('segment_id', segment_id, str)
        return Segment(self.connection, project_id, model_id).deactivate_segment(
            segment_id
        )

    def delete_segment(self, project_id: str, model_id: str, segment_id: str):
        """Deletes a deactivated segments from Fiddler"""
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        segment_id = type_enforce('segment_id', segment_id, str)
        return Segment(self.connection, project_id, model_id).delete_segment(segment_id)

    def get_segment_info(self, project_id: str, model_id: str, segment_id: str):
        """Get the SegmentInfo specified from Fiddler"""
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)
        segment_id = type_enforce('segment_id', segment_id, str)
        return Segment(self.connection, project_id, model_id).get_segment_info(
            segment_id
        )

    # End: Methods related to segment

    # def get_segment_filter_labels(self, project_id, model_id, features=None):
    #     """Potentially for use in seg-monitoring v2: custom segments"""
    #     model_info = self.get_model_info(project_id, model_id)
    #     all_cols = []
    #     if model_info.inputs is not None:
    #         all_cols += model_info.inputs
    #     if model_info.outputs is not None:
    #         all_cols += model_info.outputs
    #     if model_info.targets is not None:
    #         all_cols += model_info.targets
    #     if model_info.decisions is not None:
    #         all_cols += model_info.decisions
    #     if model_info.metadata is not None:
    #         all_cols += model_info.metadata
    #     all_symbols = [col.name for col in all_cols]
    #     if features is None:
    #         features = all_symbols
    #     invalid_cols = [col for col in features if col not in all_symbols]
    #     if len(invalid_cols) > 0:
    #         raise ValueError(f'Cannot produce segment filter labels for features {invalid_cols}, as they could not be found in the model schema.')
    #     result_features = [col for col in features]
    #     return symbols(result_features), result_features



from urllib.parse import urljoin

from fiddler.libs.http_client import RequestClient
from fiddler.utils import logging

# @todo: This is v1 implementation needs to have a proper v2 approach
from fiddler.v2.api.dataset_mixin import DatasetMixin
from fiddler.v2.api.events_mixin import EventsMixin
from fiddler.v2.api.job_mixin import JobMixin
from fiddler.v2.api.model_mixin import ModelMixin
from fiddler.v2.api.project_mixin import ProjectMixin

logger = logging.getLogger(__name__)
VERSION = '/v2'


class Client(ModelMixin, DatasetMixin, ProjectMixin, EventsMixin, JobMixin):
    def __init__(
        self, url: str, organization_name: str, auth_token: str, timeout: int = None
    ) -> None:
        self.url = urljoin(url, VERSION)
        self.auth_token = auth_token
        self.organization_name = organization_name
        self.request_headers = {'Authorization': f'Bearer {auth_token}'}
        self.timeout = timeout
        self.client = RequestClient(self.url, self.request_headers)



import os
import tarfile
import tempfile
from collections import namedtuple
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Union

import pandas as pd
import requests

from fiddler.connection import Connection
from fiddler.core_objects import (
    AttributionExplanation,
    ModelInfo,
    MulticlassAttributionExplanation,
)
from fiddler.utils import cast_input_data
from fiddler.utils.general_checks import type_enforce
from fiddler.utils.pandas import df_to_dict


class Model:
    def __init__(self, connection: Connection, project_id: str, model_id):
        self.connection = connection
        self.project_id = project_id
        self.model_id = model_id

    def get_info(self) -> ModelInfo:
        """Get ModelInfo for a model in a certain project.

        :returns: A fiddler.ModelInfo object describing the model.
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        model_id = type_enforce('model_id', self.model_id, str)

        path = ['model_info', self.connection.org_id, project_id, model_id]
        res = self.connection.call(path, is_get_request=True)
        return ModelInfo.from_dict(res)

    def delete(self, delete_prod=False, delete_pred=True):
        """Permanently delete a model.

        :param delete_prod: Boolean value to delete the production table.
            By default this table is not dropped.
        :param delete_pred: Boolean value to delete the prediction table.
            By default this table is dropped.

        :returns: Server response for deletion action.
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        model_id = type_enforce('model_id', self.model_id, str)

        payload = {
            'project_id': project_id,
            'model_id': model_id,
            'delete_prod': delete_prod,
            'delete_pred': delete_pred,
        }

        path = ['delete_model', self.connection.org_id, project_id, model_id]
        try:
            result = self.connection.call(path, json_payload=payload)
        except Exception:
            # retry on error
            result = self.connection.call(path, json_payload=payload)

        self._delete_artifacts()

        # wait for ES to come back healthy
        for i in range(3):
            try:
                self.connection.call_executor_service(
                    ['deploy', self.connection.org_id], is_get_request=True
                )
                break
            except Exception:
                pass

        return result

    def _delete_artifacts(self):
        """Permanently delete a model artifacts.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.

        :returns: Server response for deletion action.
        """
        # delete from executor service cache
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        model_id = type_enforce('model_id', self.model_id, str)

        path = ['delete_model_artifacts', self.connection.org_id, project_id, model_id]
        result = self.connection.call_executor_service(path)

        return result

    def predict(
        self,
        df: pd.DataFrame,
        log_events=False,
        casting_type=False,
    ) -> pd.DataFrame:
        """Executes a model in the Fiddler engine on a DataFrame.

        :param df: A dataframe contining model inputs as rows.
        :param log_events: Variable determining if the the predictions
            generated should be logged as production traffic
        :param casting_type: Bool indicating if fiddler should try to cast the data in the event with
        the type referenced in model info. Default to False.

        :returns: A pandas DataFrame containing the outputs of the model.
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        model_id = type_enforce('model_id', self.model_id, str)

        if casting_type:
            try:
                model_info = self.get_info()
            except RuntimeError:
                raise RuntimeError(
                    f'Did not find ModelInfo for project "{project_id}" and model "{model_id}".'
                )
            df = cast_input_data(df, model_info)

        data_array = df_to_dict(df)
        payload = dict(
            project_id=project_id,
            model_id=model_id,
            data=data_array,
            logging=log_events,
        )

        payload.pop('project_id')
        payload.pop('model_id')

        path = ['execute', self.connection.org_id, project_id, model_id]
        res = self.connection.call_executor_service(path, json_payload=payload)
        return pd.DataFrame(res)

    def explanation(
        self,
        df: pd.DataFrame,
        explanations: Union[str, Iterable[str]] = 'shap',
        dataset_id: Optional[str] = None,
        n_permutation: Optional[int] = None,
        n_background: Optional[int] = None,
        casting_type: Optional[bool] = False,
        return_raw_response=False,
    ) -> Union[
        AttributionExplanation,
        MulticlassAttributionExplanation,
        List[AttributionExplanation],
        List[MulticlassAttributionExplanation],
    ]:
        """Executes a model in the Fiddler engine on a DataFrame.

        :param df: A dataframe containing model inputs as rows. Only the first
            row will be explained.
        :param explanations: A single string or list of strings specifying
            which explanation algorithms to run.
        :param dataset_id: The unique identifier of the dataset in the
            Fiddler engine. Required for most tabular explanations, but
            optional for most text explanations.
        :param n_permutation: Number of permutations used for Fiddler SHAP. Can be used for both tabular and text data.
            By default (None), we use max(500, 2 * n_features), where n_feature is the number of word tokens
            for text input data.
        :param n_background: Number of background observations used for Fiddler SHAP for tabular data.
            By default (None), we use min(dataset.shape[0], 200)
        :param casting_type: Bool indicating if fiddler should try to cast the data in the event with
        the type referenced in model info. Default to False.

        :returns: A single AttributionExplanation if `explanations` was a
            single string, or a list of AttributionExplanation objects if
            a list of explanations was requested.
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        model_id = type_enforce('model_id', self.model_id, str)

        if casting_type:
            try:
                model_info = self.get_info()
            except RuntimeError:
                raise RuntimeError(
                    f'Did not find ModelInfo for project "{project_id}" and model "{model_id}".'
                )
            df = cast_input_data(df, model_info)

        # Explains a model's prediction on a single instance
        # wrap single explanation name in a list for the API
        if isinstance(explanations, str):
            explanations = (explanations,)

        data_array = df_to_dict(df)
        payload = dict(
            project_id=project_id,
            model_id=model_id,
            data=data_array[0],
            explanations=[dict(explanation=ex) for ex in explanations],
            n_permutation=n_permutation,
            n_background=n_background
        )
        if dataset_id is not None:
            payload['dataset'] = dataset_id

        payload.pop('project_id')
        payload.pop('model_id')

        path = ['explain', self.connection.org_id, project_id, model_id]
        res = self.connection.call_executor_service(path, json_payload=payload)

        explanations_list = res['explanations']

        if return_raw_response:
            return explanations_list

        # TODO: enable more robust check for multiclass explanations
        is_multiclass = ['explanation' not in x for x in explanations_list]
        deserialize_fn_list = [
            MulticlassAttributionExplanation.from_dict
            if x
            else AttributionExplanation.from_dict
            for x in is_multiclass
        ]

        ex_objs = [
            deserialize_fn(explanations_list[i])
            for i, deserialize_fn in enumerate(deserialize_fn_list)
        ]
        if len(ex_objs) == 1:
            return ex_objs[0]
        else:
            return ex_objs

    def feature_importance(
        self,
        dataset_id: str,
        dataset_splits: Optional[List[str]] = None,
        slice_query: Optional[str] = None,
        impact_not_importance: Optional[bool] = False,
        **kwargs,
    ) -> Dict[str, Any]:
        """Get global feature importance for a model over a dataset.

        :param dataset_id: The unique identifier of the dataset in the
            Fiddler engine.
        :param dataset_splits: If specified, importance will only be computed
            over these splits. Otherwise, all splits will be used. Only a
            single split is currently supported.
        :param slice_query: A special SQL query.
        :param impact_not_importance: Boolean flag to compute either impact or importance.
        False by default (compute feature importance).
        For text models, only feature impact is implemented.
        :param kwargs: Additional parameters to be passed to the importance
            algorithm. For example, `n_inputs`, `n_iterations`, `n_references`,
            `ci_confidence_level`.
        :return: A named tuple with the explanation results.
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        model_id = type_enforce('model_id', self.model_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)

        if (
            dataset_splits is not None
            and len(dataset_splits) > 1
            and not isinstance(dataset_splits, str)
        ):
            raise NotImplementedError(
                'Unfortunately, only a single split is '
                'currently supported for feature '
                'importances.'
            )

        source = (
            None
            if dataset_splits is None
            else dataset_splits
            if isinstance(dataset_splits, str)
            else dataset_splits[0]
        )

        payload = dict(
            subject='feature_importance',
            project_id=project_id,
            model_id=model_id,
            dataset_id=dataset_id,
            source=source,
            slice_query=slice_query,
            impact_not_importance=impact_not_importance,
        )
        payload.update(kwargs)

        payload.pop('subject')
        payload.pop('project_id')
        payload.pop('model_id')
        payload['dataset'] = payload.pop('dataset_id')

        path = ['feature_importance', self.connection.org_id, project_id, model_id]
        res = self.connection.call(path, json_payload=payload)
        # wrap results into named tuple
        res = namedtuple('FeatureImportanceResults', res)(**res)
        return res

    def fairness(
        self,
        dataset_id: str,
        protected_features: list,
        positive_outcome: Union[str, int],
        slice_query: Optional[str] = None,
        score_threshold: Optional[float] = 0.5,
    ) -> Dict[str, Any]:
        """Get fairness metrics for a model over a dataset.

        :param dataset_id: The unique identifier of the dataset in the
            Fiddler engine.
        :param protected_features: List of protected features
        :param positive_outcome: Name or value of the positive outcome
        :param slice_query: If specified, slice the data.
        :param score_threshold: positive score threshold applied to get outcomes
        :return: A dictionary with the fairness metrics, technical_metrics,
        labels distribution and model outcomes distribution
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        model_id = type_enforce('model_id', self.model_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)

        if isinstance(protected_features, str):
            protected_features = [protected_features]

        payload = dict(
            subject='fairness',
            project_id=project_id,
            model_id=model_id,
            dataset_id=dataset_id,
            protected_features=protected_features,
            slice_query=slice_query,
            score_threshold=score_threshold,
            positive_outcome=positive_outcome,
        )

        payload.pop('subject')
        payload.pop('project_id')
        payload.pop('model_id')

        path = ['fairness', self.connection.org_id, project_id, model_id]
        res = self.connection.call(path, json_payload=payload)
        return res

    def download(self, output_dir: Path):
        """
        download the model binary, package.py and model.yaml to the given
        output dir.

        :param output_dir: output directory
        :return: model artifacts
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        model_id = type_enforce('model_id', self.model_id, str)
        output_dir = type_enforce('output_dir', output_dir, Path)

        if output_dir.exists():
            raise ValueError(f'output dir already exists {output_dir}')

        headers = dict()
        headers.update(self.connection.auth_header)

        _, tfile = tempfile.mkstemp('.tar.gz')
        url = f'{self.connection.url}/get_model/{self.connection.org_id}/{project_id}/{model_id}'

        with requests.get(url, headers=headers, stream=True) as r:
            r.raise_for_status()
            with open(tfile, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    # If you have chunk encoded response uncomment if
                    # and set chunk_size parameter to None.
                    # if chunk:
                    f.write(chunk)

        tar = tarfile.open(tfile)
        output_dir.mkdir(parents=True)
        tar.extractall(path=output_dir)
        tar.close()
        os.remove(tfile)
        return True

    def _trigger_model_predictions(self, dataset_id: str):
        """Makes the Fiddler service compute and cache model predictions on a
        dataset."""
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        model_id = type_enforce('model_id', self.model_id, str)
        dataset_id = type_enforce('dataset_id', dataset_id, str)

        return self.connection.call_executor_service(
            ['dataset_predictions', self.connection.org_id, project_id],
            dict(model=model_id, dataset=dataset_id),
        )



import copy
import logging
import math
import tempfile
import threading
import time
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import deprecation
import numpy as np
import pandas as pd

from fiddler.connection import Connection
from fiddler.file_processor.src.constants import PARQUET_EXTENSION
from fiddler.project import Project

from ..aws_utils import validate_gcp_uri_access, validate_s3_uri_access
from ..core_objects import BatchPublishType, FiddlerTimestamp
from ..utils import cast_input_data
from ..utils.formatting import formatted_utcnow, print_streamed_result
from ..utils.general_checks import type_enforce
from ..utils.helper import infer_data_source
from ..utils.pandas import clean_df_types, write_dataframe_to_parquet_file

LOG = logging.getLogger(__name__)


class PublishEvent:
    def __init__(self, connection: Connection) -> None:
        self.connection = connection

    def _publish_event_batch(
        self,
        req_path: List[str],
        req_payload: Dict[str, Any],
        batch_source: Union[pd.DataFrame, str],
        data_source: Optional[BatchPublishType] = None,
        credentials: Optional[dict] = None,
    ):
        if data_source is None:
            data_source = infer_data_source(source=batch_source)
            LOG.info(
                f'`data_source` not specified. Inferred `data_source` as '
                f'BatchPublishType.{BatchPublishType(data_source.value).name}'
            )

        if data_source not in BatchPublishType:
            raise ValueError('Please specify a valid batch_source')

        assert data_source is not None, 'data_source unexpectedly None'

        req_payload['data_source'] = data_source.value

        if data_source == BatchPublishType.DATAFRAME:
            # Converting dataframe to local disk format for transmission
            req_payload['data_source'] = BatchPublishType.LOCAL_DISK.value

            assert isinstance(
                batch_source, pd.DataFrame
            ), 'batch_source unexpectedly not a DataFrame'

            df = batch_source.copy()
            if df.index.name is not None:
                df.reset_index(inplace=True)

            clean_df_types(df)

            with tempfile.TemporaryDirectory() as tmp_dir:
                file_path = Path(tmp_dir) / f'log.{PARQUET_EXTENSION}'

                write_dataframe_to_parquet_file(df=df, file_path=file_path)

                result = self.connection.call(
                    path=req_path,
                    json_payload=req_payload,
                    files=[file_path],
                    stream=True,
                )

        elif data_source == BatchPublishType.LOCAL_DISK:
            local_file_path = batch_source
            result = self.connection.call(
                path=req_path,
                json_payload=req_payload,
                files=[Path(local_file_path)],
                stream=True,
            )

        elif data_source == BatchPublishType.AWS_S3:
            s3_uri = batch_source
            # Validate S3 credentials
            validate_s3_uri_access(s3_uri, credentials, throw_error=True)
            with tempfile.NamedTemporaryFile() as tmp:
                # tmp is a dummy file passed to BE for ease of API consolidation
                req_payload['file_path'] = s3_uri
                req_payload['credentials'] = credentials
                result = self.connection.call(
                    path=req_path,
                    json_payload=req_payload,
                    files=[Path(tmp.name)],
                    stream=True,
                )

        elif data_source == BatchPublishType.GCP_STORAGE:
            gcp_uri = batch_source
            # Validate GCS credentials
            gcs_credentials: Optional[Dict[str, Any]] = None
            if credentials is not None:
                gcs_credentials = {}
                if 'gcs_access_key_id' in credentials:
                    gcs_credentials['aws_access_key_id'] = credentials[
                        'gcs_access_key_id'
                    ]
                if 'gcs_secret_access_key' in credentials:
                    gcs_credentials['aws_secret_access_key'] = credentials[
                        'gcs_secret_access_key'
                    ]
                if 'gcs_session_token' in credentials:
                    gcs_credentials['aws_session_token'] = credentials[
                        'gcs_session_token'
                    ]

            validate_gcp_uri_access(gcp_uri, gcs_credentials, throw_error=True)
            with tempfile.NamedTemporaryFile() as tmp:
                # tmp is a dummy file passed to BE for ease of API consolidation
                req_payload['file_path'] = gcp_uri
                req_payload['credentials'] = gcs_credentials
                result = self.connection.call(
                    path=req_path,
                    json_payload=req_payload,
                    files=[Path(tmp.name)],
                    stream=True,
                )
        else:
            raise ValueError(f'Data source {data_source} not supported')

        final_return = None
        for res in result:
            print_streamed_result(res)
            final_return = res

        return final_return

    def publish_events_batch(  # noqa
        self,
        project_id: str,
        model_id: str,
        batch_source: Union[pd.DataFrame, str],
        id_field: Optional[str] = None,
        update_event: Optional[bool] = False,
        timestamp_field: Optional[str] = None,
        timestamp_format: FiddlerTimestamp = FiddlerTimestamp.INFER,
        data_source: Optional[BatchPublishType] = None,
        casting_type: Optional[bool] = False,
        credentials: Optional[dict] = None,
        group_by: Optional[str] = None,
    ):
        """
        Publishes a batch events object to Fiddler Service.
        :param project_id:    The project to which the model whose events are being published belongs.
        :param model_id:      The model whose events are being published.
        :param batch_source:  Batch object to be published. Can be one of: Pandas DataFrame, CSV file, PKL Pandas DataFrame, or Parquet file.
        :param id_field:  Column to extract id value from.
        :param update_event: Bool indicating if the events are updates to previously published rows
        :param timestamp_field:     Column to extract timestamp value from.
                              Timestamp must match the specified format in `timestamp_format`.
        :param timestamp_format:   Format of timestamp within batch object. Can be one of:
                                - FiddlerTimestamp.INFER
                                - FiddlerTimestamp.EPOCH_MILLISECONDS
                                - FiddlerTimestamp.EPOCH_SECONDS
                                - FiddlerTimestamp.ISO_8601
        :param data_source:   Source of batch object. In case of failed inference, can be one of:
                                - BatchPublishType.DATAFRAME
                                - BatchPublishType.LOCAL_DISK
                                - BatchPublishType.AWS_S3
                                - BatchPublishType.GCP_STORAGE
        :param casting_type: Bool indicating if fiddler should try to cast the data in the event with
                             the type referenced in model info. Default to False.
        :param credentials:  Dictionary containing authorization for AWS or GCP.

                             For AWS S3, list of expected keys are
                              ['aws_access_key_id', 'aws_secret_access_key', 'aws_session_token']
                              with 'aws_session_token' being applicable to the AWS account being used.

                             For GCP, list of expected keys are
                              ['gcs_access_key_id', 'gcs_secret_access_key', 'gcs_session_token']
                              with 'gcs_session_token' being applicable to the GCP account being used.
        :param group_by: Column to group events together for Model Performance metrics. For example,
                         in ranking models that column should be query_id or session_id, used to
                         compute NDCG and MAP. Be aware that the batch_source file/dataset provided should have
                         events belonging to the SAME query_id/session_id TOGETHER and cannot be mixed
                         in the file. For example, having a file with rows belonging to query_id 31,31,31,2,2,31,31,31
                         would not work. Please sort the file by group_by group first to have rows with
                         the following order: query_id 31,31,31,31,31,31,2,2.
        """

        if not timestamp_format or timestamp_format not in FiddlerTimestamp:
            raise ValueError('Please specify a valid timestamp_format')

        assert timestamp_format is not None, 'timestamp_format unexpectedly None'

        path = ['publish_events_batch', self.connection.org_id, project_id, model_id]
        payload: Dict[str, Any] = {
            'casting_type': casting_type,
            'is_update_event': update_event,
            'timestamp_format': timestamp_format.value,
        }

        if id_field is not None:
            payload['id_field'] = id_field
        if timestamp_field is not None:
            payload['timestamp_field'] = timestamp_field
        if group_by is not None:
            payload['group_by'] = group_by

        # Default to current time for case when no timestamp field specified
        if timestamp_field is None:
            curr_timestamp = formatted_utcnow(
                milliseconds=int(round(time.time() * 1000))
            )
            LOG.info(
                f'`timestamp_field` not specified. Using current UTC time `{curr_timestamp}` as default'
            )
            payload['default_timestamp'] = curr_timestamp

        return self._publish_event_batch(
            req_path=path,
            req_payload=payload,
            batch_source=batch_source,
            data_source=data_source,
            credentials=credentials,
        )

    def publish_events_batch_schema(  # noqa
        self,
        batch_source: Union[pd.DataFrame, str],
        publish_schema: Dict[str, Any],
        data_source: Optional[BatchPublishType] = None,
        credentials: Optional[dict] = None,
        group_by: Optional[str] = None,
    ):
        """
        Publishes a batch events object to Fiddler Service.
        :param batch_source:  Batch object to be published. Can be one of: Pandas DataFrame, CSV file, PKL Pandas DataFrame, or Parquet file.
        :param publish_schema: Dict object specifying layout of data.
        :param data_source:   Source of batch object. In case of failed inference, can be one of:
                                - BatchPublishType.DATAFRAME
                                - BatchPublishType.LOCAL_DISK
                                - BatchPublishType.AWS_S3
                                - BatchPublishType.GCP_STORAGE
        :param credentials:  Dictionary containing authorization for AWS or GCP.

                             For AWS S3, list of expected keys are
                              ['aws_access_key_id', 'aws_secret_access_key', 'aws_session_token']
                              with 'aws_session_token' being applicable to the AWS account being used.

                             For GCP, list of expected keys are
                              ['gcs_access_key_id', 'gcs_secret_access_key', 'gcs_session_token']
                              with 'gcs_session_token' being applicable to the GCP account being used.
        :param group_by: Column to group events together for Model Performance metrics. For example,
                         in ranking models that column should be query_id or session_id, used to
                         compute NDCG and MAP.
        """

        path = ['publish_events_batch_schema', self.connection.org_id]
        payload = {
            'publish_schema': publish_schema,
            'group_by': group_by,
        }

        return self._publish_event_batch(
            req_path=path,
            req_payload=payload,
            batch_source=batch_source,
            data_source=data_source,
            credentials=credentials,
        )

    @deprecation.deprecated(
        details='WARNING: This method will be deprecated in a future version. Please use the method `publish_events_batch` instead.'
    )
    def publish_parquet_s3(
        self,
        project_id: str,
        model_id: str,
        parquet_file: str,
        auth_context: Optional[Dict[str, Any]] = None,
        ts_column: Optional[str] = None,
        default_ts: Optional[int] = None,
    ):
        """
        [SOON TO BE DEPRECATED] Publishes parquet events file from S3 to Fiddler instance. Experimental and may be expanded in the future.

        :param project_id:    The project to which the model whose events are being published belongs
        :param model_id:      The model whose events are being published
        :param parquet_file:  s3_uri for parquet file to be published
        :param auth_context:  Dictionary containing authorization for AWS. List of expected keys are
                              ['aws_access_key_id', 'aws_secret_access_key', 'aws_session_token']
                              with 'aws_session_token' being applicable to the AWS account being used.
        :param ts_column:     Column to extract time stamp value from.
                              Timestamp must be UTC in format: `%Y-%m-%d %H:%M:%S.%f` (e.g. 2020-01-01 00:00:00.000000)
                              or as the timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        :param default_ts:    Default timestamp to use if ts_column not specified.
                              Must be given as the timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        """
        # Type enforcement
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)

        payload: Dict[str, Any] = dict()
        payload['file_path'] = parquet_file
        payload['auth_context'] = auth_context

        # Validate S3 creds
        validate_s3_uri_access(parquet_file, auth_context, throw_error=True)

        if ts_column is not None:
            payload['ts_column'] = ts_column
        else:
            if default_ts is None:
                default_ts = int(round(time.time() * 1000))
            payload['default_ts'] = default_ts

        publish_path = [
            'publish_parquet_file',
            self.connection.org_id,
            project_id,
            model_id,
        ]
        publish_result = self._call(publish_path, json_payload=payload, stream=True)
        for res in publish_result:
            LOG.info(res)

    @deprecation.deprecated(
        details='WARNING: This method will be deprecated in a future version. Please use the method `publish_events_batch` instead.'
    )
    def publish_events_log(
        self,
        project_id: str,
        model_id: str,
        logs: pd.DataFrame,
        force_publish: Optional[bool] = None,
        ts_column: Optional[str] = None,
        default_ts: Optional[int] = None,
        num_threads: int = 1,
        batch_size: Optional[int] = None,
        casting_type: Optional[bool] = False,
    ):
        """
        [SOON TO BE DEPRECATED] Publishes prediction log to Fiddler Service.
        :param project_id:    The project to which the model whose events are being published belongs
        :param model_id:      The model whose events are being published
        :param logs:          Data frame of event details, such as features and predictions.
        :param force_publish: Continue with publish even if all input and output columns not in log.
        :param ts_column:     Column to extract timestamp value from.
                              Timestamp must be UTC in format: `%Y-%m-%d %H:%M:%S.%f` (e.g. 2020-01-01 00:00:00.000000)
                              or as the timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        :param default_ts:    Default timestamp to use if ts_column not specified.
                              Must be given as the timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        :param num_threads:   Number of threads to parallelize this function over. Dataset will be divided evenly
                              over the number of threads.
        :param batch_size:    Size of dataframe to publish in any given run.
        :param casting_type: Bool indicating if fiddler should try to cast the data in the event with
        the type referenced in model info. Default to False.

        """
        project_id = type_enforce('project_id', project_id, str)
        model_id = type_enforce('model_id', model_id, str)

        if num_threads < 1 or num_threads > 20:
            raise ValueError(
                'Please adjust parameter `num_threads` to be between 1 and 20.'
            )
        if batch_size and (batch_size < 1 or batch_size > 100000):
            raise ValueError(
                'Please adjust parameter `batch_size` to be between 1 and 100000, or type `None`.'
            )
        if default_ts and not isinstance(default_ts, int):
            raise ValueError(
                'Please adjust parameter `default_ts` to be of type `int` or type `None`.'
            )
        model_info = Project(self.connection, project_id).model(model_id).get_info()
        # model_info = self.get_model_info(project_id, model_id)
        if casting_type:
            logs = cast_input_data(logs, model_info)

        log_columns = [c for c in list(logs.columns)]
        in_columns = [c.name for c in model_info.inputs]
        in_not_found = [c for c in in_columns if c not in log_columns]
        out_columns = [c.name for c in model_info.outputs]
        out_not_found = [c for c in out_columns if c not in log_columns]
        if (out_not_found or in_not_found) and not force_publish:
            raise ValueError(
                f'Model output columns "{out_not_found}" or input columns "{in_not_found}"'
                f'not found in logs. If this is expected try again with force_publish=True.'
            )

        payload: Dict[str, Any] = dict()
        if ts_column is not None:
            if ts_column not in log_columns:
                raise ValueError(f'Did not find {ts_column} in the logs columns.')
            payload['ts_column'] = ts_column
        else:
            if default_ts is None:
                default_ts = int(round(time.time() * 1000))
            payload['default_ts'] = default_ts
        include_index = logs.index.name is not None

        worker_lock = threading.Lock()
        workers_list = []

        class _PublishEventsLogWorker(threading.Thread):
            """
            Handles the call to `publish_events_log`
            """

            def __init__(
                self,
                thread_id,
                df,
                client,
                payload,
                org_id,
                project_id,
                model_id,
                include_index,
                batch_size,
                worker_lock,
                only_worker,
            ):
                threading.Thread.__init__(self)
                self.thread_id = thread_id
                self.df = df
                self.client = client
                self.payload = copy.deepcopy(payload)
                self.org_id = org_id
                self.project_id = project_id
                self.model_id = model_id
                self.include_index = include_index
                self.batch_size = batch_size
                self.worker_lock = worker_lock
                self.only_worker = only_worker
                self.path = [
                    'publish_events_log',
                    self.org_id,
                    self.project_id,
                    self.model_id,
                ]

            def run(self):
                df_batches = []
                if self.batch_size:
                    # Divide dataframe into size of self.batch_size
                    num_chunks = math.ceil(len(self.df) / self.batch_size)
                    for j in range(num_chunks):
                        df_batches.append(
                            self.df[j * self.batch_size : (j + 1) * self.batch_size]
                        )
                else:
                    df_batches.append(self.df)

                with tempfile.TemporaryDirectory() as tmp:
                    # To maintain the same data types during transportation from client
                    #  to server, we must explicitly send and recreate the data types through
                    #  a file. Otherwise, Pandas.from_csv() will convert quoted strings to integers.
                    #  See: https://github.com/pandas-dev/pandas/issues/35713
                    log_path = Path(tmp) / 'log.csv'
                    dtypes_path = Path(tmp) / 'dtypes.csv'
                    self.df.dtypes.to_frame('types').to_csv(dtypes_path)
                    csv_paths = [dtypes_path, log_path]

                    for curr_batch in df_batches:
                        # Overwrite CSV with current batch
                        curr_batch.to_csv(log_path, index=self.include_index)

                        result = self.client._call(
                            self.path, json_payload=self.payload, files=csv_paths
                        )

                        with self.worker_lock:
                            # Used to prevent printing clash
                            if self.only_worker:
                                LOG.info(result)
                            else:
                                LOG.info(f'thread_id {self.thread_id}: {result}')

        # Divide dataframe evenly amongst each thread
        df_split = np.array_split(logs, num_threads)

        for i in range(num_threads):
            workers_list.append(
                _PublishEventsLogWorker(
                    i,
                    df_split[i],
                    self,
                    payload,
                    self.connection.org_id,
                    project_id,
                    model_id,
                    include_index,
                    batch_size,
                    worker_lock,
                    num_threads == 1,
                )
            )
            workers_list[i].start()

        for i in range(num_threads):
            workers_list[i].join()



import os
import tempfile
from abc import ABC, abstractmethod

from fiddler.file_processor.src.constants import (
    FILE_EXTENSIONS_TO_CODE,
    SUPPORTABLE_FILE_EXTENSIONS,
    ExtractorDataType,
)


class FileExtractor(ABC):
    @staticmethod
    def get_extractor(datasource):
        if datasource == 'LOCAL_DISK':
            return LocalFileExtractor()
        if datasource == 'S3':
            return S3FileExtractor()
        if datasource == 'GCP':
            return GCPFileExtractor()

    @staticmethod
    def get_supportable_file_extension(object_path, fast_fail=True):
        """
        Given an `object_path`, will return the file's extension (e.g. ".parquet", ".csv")

        If `fast_fail`, will throw an exception should not able to find a supportable file extension
        """
        file_extension = os.path.splitext(object_path)[1]
        for supported_file_extension in SUPPORTABLE_FILE_EXTENSIONS:
            if file_extension in supported_file_extension:
                return FILE_EXTENSIONS_TO_CODE[file_extension]

        if fast_fail:
            err_val = f'Error: File extension not currently supported. \
                        Supported formats are : {SUPPORTABLE_FILE_EXTENSIONS} \
                        Provided format is :{file_extension}'
            raise ValueError(err_val)
        else:
            return ExtractorDataType.UNKNOWN

    @abstractmethod
    def extract(files):
        pass

    def unzip(self):
        pass


class LocalFileExtractor(FileExtractor):
    def extract(self, files):
        """
        Downloads files sent from Fiddler client's local disk to the current pod's disk,
        allowing for further processing.
        """
        temp_dir = tempfile.TemporaryDirectory()
        file_list = []
        for name, file in files.items():
            file_extension = self.get_supportable_file_extension(file.filename)
            file_path = os.path.join(temp_dir.name, file.filename)
            file.save(file_path)
            file_list.append((temp_dir, file_path, file_extension))
        return file_list


class S3FileExtractor(FileExtractor):
    pass


class GCPFileExtractor(FileExtractor):
    pass



from fiddler.file_processor.src.extractor import FileExtractor
from fiddler.file_processor.src.file_workflow_executor import (
    DatasetUploadWorkflowExecutor,
)
from fiddler.file_processor.src.processor import FileProcessorBase


def upload_dataset(files, source_data_source, file_type, schema=None):
    extractor = FileExtractor.get_extractor(source_data_source)
    processor = FileProcessorBase.get_processor(file_type, schema)
    executor = DatasetUploadWorkflowExecutor(extractor, processor)
    return executor.execute(files)



DATASET_FIDDLER_ID = '__fiddler_id'
ONE_LINE_PRINT = '¡í€'
TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
MAX_ID_LEN = 30


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDCOLOR = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'



import base64
import csv
import json
import logging
import uuid
from io import StringIO
from os import path
from urllib.parse import urlparse

import boto3
import botocore

bucket = 'sagemaker-us-west-2-079310353266'
prefix = 'sagemaker/sagemaker-xgboost-lending/'
s3_client = boto3.client('s3')

LOGGER = logging.getLogger(__name__)


class S3Uri:
    """bucket, object_key for an S3 uri."""

    def __init__(
        self, bucket: str, key: str, auth_context: dict = {}, endpoint_url=None
    ):
        self.bucket = bucket
        self.key = key

        if auth_context and (
            auth_context.get('aws_access_key_id')
            and auth_context.get('aws_secret_access_key')
        ):
            # using supplied credentials
            self.s3 = boto3.resource(
                's3',
                endpoint_url=endpoint_url,
                aws_access_key_id=auth_context['aws_access_key_id'],
                aws_secret_access_key=auth_context['aws_secret_access_key'],
                aws_session_token=auth_context.get('aws_session_token'),
            )
        elif boto3.session.Session().get_credentials():
            # using default credentials
            # if not auth_context, assume that the credentials are set up
            # elsewhere viz., envvars, credentials file etc.
            LOGGER.info('Gettting credentials from the boto3 session')
            self.s3 = boto3.resource('s3', endpoint_url=endpoint_url)
        else:
            # there are no credentials.. Maybe, the user is expecting to
            # download from a public S3 bucket?
            # using no credentials
            LOGGER.info('No credentials provided using public bucket')
            cnfg = botocore.config.Config(signature_version=botocore.UNSIGNED)
            self.s3 = boto3.resource('s3', endpoint_url=endpoint_url, config=cnfg)

    @property
    def uri(self):
        return f's3://{self.bucket}/{self.key}'

    def s3_bucket(self):  # Bucket
        """Returns Bucket object `boto3.resource('s3').Bucket(bucket)`."""
        return self.s3.Bucket(self.bucket)

    def s3_object(self, relative_path: str = None):  # Bucket.Object()
        """Returns S3 Object for path under this uri."""
        if not relative_path:
            return self.s3_bucket().Object(self.key)
        else:
            return self.s3_bucket().Object(path.join(self.key, relative_path))

    def read_object(self, relative_path: str = None) -> bytes:
        """Reads contents of object located at `relative_path`.
        Be careful not to read large objects.
        :returns: file content as byte array"""
        return self.s3_object(relative_path).get()['Body'].read()

    def is_object_readable(self, relative_path: str = None):
        """Reads object located at `relative_path`.
        :returns: True if successful."""
        try:
            self.s3_object(relative_path).get()
        except botocore.exceptions.ClientError as e:
            return False, str(e)
        return True, None

    def download_file(self, local_path: str, relative_path: str = None):
        """Downloads the remote file `relative_path` to `local_path`."""
        return self.s3_object(relative_path).download_file(local_path)

    @staticmethod
    def from_uri(s3_uri: str, auth_context: dict = {}):
        parsed = urlparse(s3_uri)
        if parsed.scheme != 's3':
            raise ValueError(f'`s3://<bucket>/<prefix>` expected, but got `{s3_uri}`')
        return S3Uri(parsed.netloc, parsed.path.lstrip('/'), auth_context)


class GcsUri(S3Uri):
    """bucket, object_key for an S3 uri."""

    def __init__(self, bucket: str, key: str, auth_context: dict = {}):
        endpoint_url = 'https://storage.googleapis.com'
        super().__init__(bucket, key, auth_context, endpoint_url=endpoint_url)

    @staticmethod
    def from_uri(gs_uri: str, auth_context: dict = {}):
        parsed = urlparse(gs_uri)
        if parsed.scheme != 'gs':
            raise ValueError(f'`gs://<bucket>/<prefix>` expected, but got `{gs_uri}`')
        return S3Uri(
            parsed.netloc,
            parsed.path.lstrip('/'),
            auth_context,
            endpoint_url='https://storage.googleapis.com',
        )


def validate_s3_uri_access(url, credentials, throw_error=False):
    if credentials is None:
        credentials = {}
    s3_handle = S3Uri.from_uri(url, credentials)
    is_valid_s3, message = s3_handle.is_object_readable()
    if not is_valid_s3:
        if throw_error:
            raise ValueError(f'Unable to access S3 URI\n{message}')
        return False

    return True


def validate_gcp_uri_access(url, credentials, throw_error=False):
    if credentials is None:
        credentials = {}
    gcs_handle = GcsUri.from_uri(url, credentials)
    is_valid_gcs, message = gcs_handle.is_object_readable()
    if not is_valid_gcs:
        if throw_error:
            raise ValueError(f'Unable to access GS URI\n{message}')
        return False

    return True


def sample_data_capture_log(bucket, prefix, sample_size):
    objects = s3_client.list_objects(Bucket=bucket, Prefix=prefix)

    result = []
    for obj in objects['Contents']:
        key = obj['Key']
        if key.endswith('.jsonl'):
            tmpkey = key.replace('/', '')
            download_path = '/tmp/{}{}'.format(uuid.uuid4(), tmpkey)
            s3_client.download_file(bucket, key, download_path)
            with open(download_path) as f:
                for line in f:
                    pline = json.loads(line)
                    input = base64.b64decode(
                        pline['captureData']['endpointInput']['data']
                    ).decode('ascii')
                    output = base64.b64decode(
                        pline['captureData']['endpointOutput']['data']
                    ).decode('ascii')
                    # input = pline['captureData']['endpointInput']['data']
                    # output = pline['captureData']['endpointOutput']['data']
                    inputstr = StringIO(input)
                    inarray = next(csv.reader(inputstr, delimiter=','))
                    outputstr = StringIO(output)
                    outarray = next(csv.reader(outputstr, delimiter=','))
                    result.append(inarray + outarray)
                    if len(result) >= sample_size:
                        return result

    return result



from datetime import datetime
from typing import List

from . import constants
from .core_objects import (
    DataType,
    EventTypes,
    FiddlerEventColumns,
    ModelInputType,
    ModelTask,
    MonitoringViolation,
    MonitoringViolationType,
)
from .utils.formatting import formatted_utcnow, pad_timestamp
from .utils.general_checks import is_int_type


class MonitoringValidator:
    def __init__(self):
        pass

    def pre_flight_monitoring_check(
        self, event, model_info, dataset_info
    ) -> List[MonitoringViolation]:

        violations = []
        # lets track all violations we can find. Its much better to find everything
        # wrong with the setup rather than error-ing out on each step and iterating
        # We only return immidiately if we have a FATAL violation

        violations += self._input_types_allowed(model_info)
        violations += self._basic_outputs_checks(model_info)

        violations += self._basic_decision_checks(model_info)
        violations += self._basic_monitoring_single_event_tests(
            dataset_info, model_info, event
        )
        violations += self._basic_data_integrity_checks(dataset_info)
        violations += self._basic_accuracy_checks(model_info)

        return violations

    def _input_types_allowed(self, model_info):
        violations = []
        string_columns = []
        for col in model_info.inputs:
            if (
                col.data_type != DataType.FLOAT
                and col.data_type != DataType.INTEGER
                and col.data_type != DataType.CATEGORY
                and col.data_type != DataType.BOOLEAN
            ):
                if col.data_type == DataType.STRING:
                    string_columns.append(col.name)
                else:
                    msg = f'Column {col.name} of type {col.data_type} is not allowed'
                    violations.append(
                        MonitoringViolation(MonitoringViolationType.FATAL, msg)
                    )
        if string_columns:
            msg = f'Feature of type {DataType.STRING} are not monitored currently. {DataType.STRING} columns in the event are {string_columns}'
            violations.append(MonitoringViolation(MonitoringViolationType.WARNING, msg))
        return violations

    def _check_missing_inputs_outputs(self, dataset_info, model_info, event):
        """Lets make sure we get all not is_nullable features. Also, the outputs
        SHOULD be there in execution type event.
        """
        violations = []
        # We have to check nullability from the dataset info which is what the DI
        # checks are based on.
        input_col_types = {col.name: col for col in dataset_info.columns}
        for col in model_info.inputs:
            if (
                col.name not in event
                and input_col_types[col.name].is_nullable is not None
                and input_col_types[col.name].is_nullable is False
            ) and event[
                FiddlerEventColumns.EVENT_TYPE.value
            ] == EventTypes.EXECUTION_EVENT.value:
                msg = f'Input is_nullable error: `{col.name}` can not be missing'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.WARNING, msg)
                )

        for col in model_info.outputs:
            if (
                col.name not in event
                and event[FiddlerEventColumns.EVENT_TYPE.value]
                == EventTypes.EXECUTION_EVENT.value
            ):

                msg = f'Output error: `{col.name}` can not be missing'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.WARNING, msg)
                )

        return violations

    def _event_type_based_checks(self, model_info, event):
        violations = []
        if FiddlerEventColumns.EVENT_TYPE.value not in event:
            msg = 'Can not ingest event with missing event_type'
            violations.append(MonitoringViolation(MonitoringViolationType.FATAL, msg))
            return violations

        if event[FiddlerEventColumns.EVENT_TYPE.value] == EventTypes.UPDATE_EVENT.value:

            # We dont accept input/output with event_type = 'update_event'
            for col in model_info.inputs:
                if col.name in event.keys():
                    msg = f'Can not send inputs/features in update_event {col.name}'
                    violations.append(
                        MonitoringViolation(MonitoringViolationType.FATAL, msg)
                    )

            for col in model_info.outputs:
                if col.name in event.keys():
                    msg = f'Can not send outputs in update_event {col.name}'
                    violations.append(
                        MonitoringViolation(MonitoringViolationType.FATAL, msg)
                    )

        return violations

    def _check_event_timestamps(self, event):
        violations = []
        date_str = None
        # We are checking to see if event_type is present. Dont want to rely on event_type
        # being present.
        if (
            FiddlerEventColumns.EVENT_TYPE.value in event
            and event[FiddlerEventColumns.EVENT_TYPE.value]
            == EventTypes.EXECUTION_EVENT.value
        ):
            # __occurred_at has to be present for execution_event
            if (
                FiddlerEventColumns.OCCURRED_AT.value not in event
                or event[FiddlerEventColumns.OCCURRED_AT.value] is None
            ):
                msg = f'execution_events need to have `{FiddlerEventColumns.OCCURRED_AT.value}` included in the event'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.FATAL, msg)
                )
            else:
                date_str = event[FiddlerEventColumns.OCCURRED_AT.value]

        if (
            FiddlerEventColumns.EVENT_TYPE.value in event
            and event[FiddlerEventColumns.EVENT_TYPE.value]
            == EventTypes.UPDATE_EVENT.value
        ):
            # __updated_at should be present for update_event
            if (
                FiddlerEventColumns.UPDATED_AT.value not in event
                or event[FiddlerEventColumns.UPDATED_AT.value] is None
            ):
                msg = f'update_events need to have `{FiddlerEventColumns.UPDATED_AT.value}` included in the event'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.FATAL, msg)
                )
            else:
                date_str = event[FiddlerEventColumns.UPDATED_AT.value]

        if date_str is None:
            return violations

        # We expect either date to be in '%Y-%m-%d %H:%M:%S.%f' format or millisecs epoch.
        # Some can potentially send us date in millisecs for 1970. So its worth pointing
        # out if they intend to do that or its just a secs/millisecs problem.
        is_int, int_val = is_int_type(date_str)
        if is_int:
            t_stamp = formatted_utcnow(milliseconds=int_val)
            year = int(t_stamp.split('-')[0])
            if year < 1975:
                msg = f'Is your timestamp ({int_val}) in secs? Please change it to millisecs.'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.WARNING, msg)
                )
        elif isinstance(date_str, str):
            try:
                ts = pad_timestamp(date_str)
                datetime.strptime(ts, constants.TIMESTAMP_FORMAT)
            except Exception:
                # lets check if its valid ms epoch time
                is_int, int_val = is_int_type(ts)
                msg = f'Unable to parse date:{date_str}. Specify date either in millisecs epoch or format:{constants.TIMESTAMP_FORMAT}'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.FATAL, msg)
                )

        return violations

    def _basic_monitoring_single_event_tests(self, dataset_info, model_info, event):
        """These checks would currently apply to only publish_event function and not
        publish_events_log or parquet function. We can potentially convert the df row
        from publish_events_log to json and use this same function.
        However, that might not be the best way since we are not doing that transform on
        server side before ingestion. Once we have the ETL work done, we should figure
        out how to do a single transform to all incoming events (whether publish_events_log
        or publish_event) and use the same transform here before calling this function.
        """
        violations = []
        if event is None or not event:
            msg = 'No event passed'
            violations.append(MonitoringViolation(MonitoringViolationType.WARNING, msg))
            return violations

        violations += self._check_event_timestamps(event)

        violations += self._event_type_based_checks(model_info, event)

        all_cols = model_info.inputs + model_info.outputs + model_info.targets

        if model_info.decisions:
            all_cols += model_info.decisions
        if model_info.metadata:
            all_cols += model_info.metadata

        model_column_type = {}
        for col in all_cols:
            model_column_type[col.name] = col.data_type

        input_col_types = {col.name: col.data_type for col in model_info.inputs}
        fiddler_col_names = [col.value for col in FiddlerEventColumns]

        violations += self._check_missing_inputs_outputs(
            dataset_info, model_info, event
        )

        # lets check if dtype in events match model input schema
        for col_name, value in event.items():
            if col_name in fiddler_col_names:
                continue

            if col_name not in input_col_types.keys():
                continue

            value_dtype = type(value).__name__
            if input_col_types[col_name].is_numeric():
                if value_dtype == 'int' or value_dtype == 'float':
                    continue
                msg = f'Input dtype violation for {col_name}: Found {value_dtype} expected {input_col_types[col_name].value}'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.WARNING, msg)
                )
            elif input_col_types[col_name] == DataType.CATEGORY:
                if value_dtype != 'str':
                    msg = f'Input dtype violation for {col_name}: Found {value_dtype} expected category (string).'
                    violations.append(
                        MonitoringViolation(MonitoringViolationType.WARNING, msg)
                    )
            else:
                if value_dtype != input_col_types[col_name].value:
                    msg = f'Input dtype violation for {col_name}: Found {value_dtype} expected {input_col_types[col_name].value}'
                    violations.append(
                        MonitoringViolation(MonitoringViolationType.WARNING, msg)
                    )

        # lets check if dtype in events match model output schema
        output_col_types = {col.name: col.data_type for col in model_info.outputs}
        for col_name, value in event.items():

            if col_name in fiddler_col_names:
                continue

            if col_name not in output_col_types.keys():
                continue

            value_dtype = type(value).__name__
            if value_dtype != output_col_types[col_name].value:
                msg = f'Output dtype violation for {col_name}: Found {value_dtype} expected {output_col_types[col_name].value}'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.WARNING, msg)
                )

        return violations

    def _basic_decision_checks(self, model_info):
        # TODO: add a check for decision in the event matching dtype mentioned in the model
        violations = []
        if model_info.decisions:
            for col in model_info.decisions:
                if col.data_type != DataType.CATEGORY:
                    msg = 'Decision column can only be of type category'
                    violations.append(
                        MonitoringViolation(MonitoringViolationType.WARNING, msg)
                    )
        return violations

    def _basic_outputs_checks(self, model_info):
        violations = []
        for col in model_info.outputs:
            if not col.data_type.is_numeric():
                msg = 'Model outputs can ONLY be numeric'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.WARNING, msg)
                )
        return violations

    def _basic_data_integrity_checks(self, dataset_info):
        # Lets make sure dataset_info has stats present
        violations = []
        for col in dataset_info.columns:
            if col.is_nullable is None or not hasattr(col, 'is_nullable'):
                msg = f'Data integrity error: {col.name} doesnt have is_nullable'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.WARNING, msg)
                )

            if col.data_type.is_numeric():
                if col.value_range_min is None or not hasattr(col, 'value_range_min'):
                    msg = (
                        f'Data integrity error: {col.name} doesnt have value_range_min'
                    )
                    violations.append(
                        MonitoringViolation(MonitoringViolationType.WARNING, msg)
                    )

                if col.value_range_max is None or not hasattr(col, 'value_range_max'):
                    msg = (
                        f'Data integrity error: {col.name} doesnt have value_range_max'
                    )
                    violations.append(
                        MonitoringViolation(MonitoringViolationType.WARNING, msg)
                    )

        return violations

    def _basic_accuracy_checks(self, model_info):
        violations = []

        # We currently only support Tabular data for monitoring
        if model_info.input_type != ModelInputType.TABULAR:
            msg = f'Model:{model_info.display_name} type:{model_info.input_type} is not supported. Monitoring currently only supports {ModelInputType.TABULAR.value}'
            violations.append(MonitoringViolation(MonitoringViolationType.FATAL, msg))

        for col in model_info.outputs:
            if col.data_type != DataType.FLOAT:
                msg = f'Accuracy metrics can only be computed for models which have floating point output columns:{col.name}'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.WARNING, msg)
                )

        if model_info.model_task == ModelTask.BINARY_CLASSIFICATION:
            if model_info.targets[0].data_type not in [
                DataType.FLOAT,
                DataType.INTEGER,
                DataType.BOOLEAN,
                DataType.CATEGORY,
            ]:
                msg = f'Target columns can only be one of `FLOAT`, `INTEGER`, `BOOLEAN`, or `CATEGORY` for model type:{ModelTask.BINARY_CLASSIFICATION.value}'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.WARNING, msg)
                )
            if model_info.targets[0].possible_values is not None:
                for value in model_info.targets[0].possible_values:
                    if (
                        not isinstance(value, str)
                        and not isinstance(value, float)
                        and not isinstance(value, bool)
                        and not isinstance(value, int)
                    ):
                        msg = f'Accuracy only works if possible values for target/label columns are strings, floats, integers, or booleans. Current target possible values are {model_info.targets[0].possible_values}'
                        violations.append(
                            MonitoringViolation(MonitoringViolationType.WARNING, msg)
                        )
                        # no need at add a warning for each possible value
                        break

        if model_info.model_task == ModelTask.MULTICLASS_CLASSIFICATION:
            if model_info.targets[0].data_type != DataType.CATEGORY:
                msg = f'Target columns can only be categories for model type:{model_info.model_task.value}'
                violations.append(
                    MonitoringViolation(MonitoringViolationType.WARNING, msg)
                )
            if model_info.targets[0].possible_values is not None:
                msg = f'Accuracy will only work if the targets for Multiclass model can point to the winning output probability. For your case, the target possible values should be {list(range(len(model_info.outputs)))}'
                # lets make sure that possible values cover the range of outputs
                # If there are three outputs probs, then target possible values = [0, 1, 2]
                int_target_vals = None
                try:
                    int_target_vals = set(
                        [int(c) for c in model_info.targets[0].possible_values]
                    )
                except ValueError:
                    violations.append(
                        MonitoringViolation(MonitoringViolationType.WARNING, msg)
                    )
                if int_target_vals is not None and not (
                    len(int_target_vals) == len(model_info.outputs)
                    and set(int_target_vals) == set(range(len(model_info.outputs)))
                ):
                    violations.append(
                        MonitoringViolation(MonitoringViolationType.WARNING, msg)
                    )

        return violations



import os
from contextlib import contextmanager
from os import listdir
from os.path import join

from fiddler import constants


def file_list(root_dir):
    result = []
    if not os.path.isdir(root_dir):
        return result

    for f in listdir(root_dir):
        file_stats = os.stat(join(root_dir, f))
        result.append(
            {'name': f, 'size': file_stats.st_size, 'modified': file_stats.st_mtime}
        )
    return result


def getProgressBar(
    iteration, total, prefix='', suffix='', decimals=1, length=50, fill='█'
):
    """
    Adapted with love from https://stackoverflow.com/a/34325723

    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    BAR_COLOR = constants.bcolors.OKBLUE
    EMPTY_CHARACTER = '-'
    percent = ('{0:.' + str(decimals) + 'f}').format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + EMPTY_CHARACTER * (length - filledLength)
    return f'{constants.ONE_LINE_PRINT}{prefix} |{BAR_COLOR}{bar}{constants.bcolors.ENDCOLOR}| {percent}% {suffix}'


@contextmanager
def open_file_w_auto_close(filename, mode='rb'):
    f = open(filename, mode=mode)
    try:
        yield f
    finally:
        f.close()



from fiddler.core_objects import (
    DatasetInfo,
    DataType,
    ModelInfo,
    ModelInputType,
    ModelTask,
)


class ModelInfoValidator:
    def __init__(self, model_info: ModelInfo, dataset_info: DatasetInfo):
        self.model_info = model_info
        self.dataset_info = dataset_info

    def validate_categoricals(self, modify=False):
        for col in self.model_info.inputs:
            possible_values_floats = None
            if col.data_type == DataType.CATEGORY.value:
                try:
                    possible_values_floats = [
                        str(float(raw_val)) for raw_val in col.possible_values
                    ]
                except ValueError:
                    pass
            if possible_values_floats is not None:
                if not modify:
                    for f_val, c_val in zip(
                        possible_values_floats, col.possible_values
                    ):
                        if str(f_val) != str(c_val):
                            raise ValueError(
                                f'Categoricals have invalid values: {f_val} != {c_val}'
                            )
                else:
                    col.possible_values = possible_values_floats

    def validate(self):
        if self.model_info.input_type != ModelInputType.TABULAR:
            raise ValueError('Only tabular models supported')

        if not self.model_info.targets:
            raise ValueError('Target not specified')

        if len(self.model_info.targets) != 1:
            raise ValueError('More than one target specified')

        if len(self.model_info.inputs) < 1:
            raise ValueError('Input features not specified')

        if len(self.model_info.outputs) < 1:
            raise ValueError('Model outputs not defined')

        if self.model_info.model_task == ModelTask.REGRESSION:
            self.validate_regression_model()
        elif self.model_info.model_task == ModelTask.BINARY_CLASSIFICATION:
            self.validate_binary_classification_model()
        elif self.model_info.model_task == ModelTask.MULTICLASS_CLASSIFICATION:
            self.validate_multiclass_classification()
        elif self.model_info.model_task == ModelTask.RANKING:
            self.validate_ranking()
        else:
            raise ValueError('unsupported model task')

    def validate_regression_model(self):
        if len(self.model_info.outputs) != 1:
            raise ValueError('only one model output can be specified')
        if self.model_info.outputs[0].data_type != DataType.FLOAT:
            raise ValueError('model output must be of type FLOAT')

    def validate_binary_classification_model(self):
        if len(self.model_info.outputs) != 1:
            raise ValueError('only one model output can be specified')
        if self.model_info.outputs[0].data_type != DataType.FLOAT:
            raise ValueError('model output must be of type FLOAT')
        if self.model_info.targets[0].data_type == DataType.STRING:
            raise ValueError(
                'Target cannot be of type string. Please review dataset schema to be type category.'
            )
        if (
            self.model_info.targets[0].data_type
            in [DataType.CATEGORY, DataType.BOOLEAN]
        ) and (len(self.model_info.targets[0].possible_values) != 2):
            raise ValueError('target must have two possible values')

    def validate_multiclass_classification(self):
        if len(self.model_info.outputs) < 1:
            raise ValueError('model output should be more than one')
        for item in self.model_info.outputs:
            if item.data_type != DataType.FLOAT:
                raise ValueError(
                    f'model output "{item.name}" ' f'must be of type FLOAT'
                )
        if self.model_info.targets[0].data_type == DataType.STRING:
            raise ValueError(
                'Target cannot be of type string. Please review dataset schema to be type category.'
            )
        if (
            self.model_info.targets[0].data_type
            in [DataType.CATEGORY, DataType.BOOLEAN]
        ) and (
            len(self.model_info.targets[0].possible_values)
            != len(self.model_info.outputs)
        ):
            raise ValueError(
                f'possible values in target does not match model '
                f'outputs target: {self.model_info.targets[0]} '
                f'outputs: {self.model_info.outputs} '
            )

    def validate_ranking(self):
        if len(self.model_info.outputs) != 1:
            raise ValueError('only one model output can be specified')
        if self.model_info.outputs[0].data_type != DataType.FLOAT:
            raise ValueError('model output must be of type FLOAT')
        if self.model_info.targets[0].data_type == DataType.STRING:
            raise ValueError(
                'Target cannot be of type string. Please review dataset schema.'
            )

    def validate_weighting(self):
        # TODO: validate weighting params
        pass



from abc import ABC, abstractmethod
from dataclasses import dataclass

from fiddler.file_processor.src.extractor import FileExtractor
from fiddler.file_processor.src.processor import FileProcessorBase


class WorkflowExecutor(ABC):
    @abstractmethod
    def execute(self, files):
        pass


@dataclass
class DatasetUploadWorkflowExecutor(WorkflowExecutor):
    extractor: FileExtractor
    processor: FileProcessorBase

    def execute(self, files):
        files_list = self.extractor.extract(files)
        return self.processor.process(files_list)



import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from fiddler.file_processor.src.constants import (
    PARQUET_COMPRESSION,
    PARQUET_ROW_GROUP_SIZE,
)

from .. import constants

LOG = logging.getLogger(__name__)


def is_datetime(pandas_series):
    """
    Return true if Series contains all valid dates.
    """
    # if pandas_series contains valid timestamp series to_datetime will not
    # throw an exception, which means all rows has timestamp.
    try:
        pd.to_datetime(pandas_series, format=constants.TIMESTAMP_FORMAT, errors='raise')
        return True
    except Exception:
        return False


def clean_df_types(df):
    """
    Cleans the dataframe types into serializable formats where needed. Currently
    works to:
    - Convert datetime to string

    TODO: Expand to all other datatypes
    """
    for ind, col_type in enumerate(df.dtypes):
        col = df.columns[ind]
        if (
            col_type is not None
            and 'datetime64' in str(col_type)
        ):
            df[col] = pd.to_datetime(
                df[col], utc=True
            ).dt.tz_localize(None)  # Skipping timezone information from the column. REF: https://fiddlerlabs.atlassian.net/browse/FDL-2592
            df[col] = df[col].dt.strftime(constants.TIMESTAMP_FORMAT)
            df[col] = df[col].astype('string')

        # pandas considers the datatype as an object, if it finds multiple datatypes in a column
        # If such columns are consists of timestamp values, and those values are less than
        # specified 'max_inferred_cardinality' value. System infers the timestamp variable as a categorical
        # variable. To avoid that, we added a check where we check the object is a datetime object
        # or not. If yes, we convert the datetime object into string.
        elif col_type == object:
            if df[col].astype('string') is not None and is_datetime(
                df[col].astype('string')
            ):
                df[col] = pd.to_datetime(
                    df[col], format=constants.TIMESTAMP_FORMAT, utc=True
                ).dt.tz_localize(None)
                df[col] = df[col].astype('string')
    return df


def try_series_retype(series: pd.Series, new_type) -> Union[pd.DataFrame, pd.Series]:
    try:
        return series.astype(new_type)
    except (TypeError, ValueError) as e:
        if new_type == 'int':
            LOG.warning(
                f'"{series.name}" cannot be loaded as int '
                f'(likely because it contains missing values, and '
                f'Pandas does not support NaN for ints). Loading '
                f'as float instead.'
            )
            return series.astype('float')
        elif new_type == 'TIMESTAMP':
            try:
                return series.apply(
                    lambda x: datetime.strptime(x, constants.TIMESTAMP_FORMAT)
                )
            # if timestamp str doesn't contain millisec.
            # @NOTE: This can be a makeshift fix.
            except ValueError:
                # @TODO: Should such cases be
                # 1. handled by client OR
                # 2. should server apped 00 milliseconds if not present during ingestion OR
                # 3. should it break if timestamp format does not match while ingestion?
                return series.apply(lambda x: datetime.strptime(x, '%Y-%m-%d %H:%M:%S'))
        else:
            raise e


def df_to_dict(df: pd.DataFrame):
    data_array = [y.iloc[0, :].to_dict() for x, y in df.groupby(level=0)]  # type: ignore
    # convert numpy type values to python type: some numpy types are not JSON serializable
    for data in data_array:
        for key, val in data.items():
            if isinstance(val, np.bool_):
                data[key] = bool(val)
            if isinstance(val, np.integer):
                data[key] = int(val)
            if isinstance(val, np.floating):
                data[key] = float(val)

    return data_array


def df_size_exceeds(dataset: Dict[str, pd.DataFrame], max_len: int):
    """
    Returns True if size of any of the dataframes exceeds max_len
    """
    for name, df in dataset.items():
        if df.shape[0] > max_len:
            return True
    return False


def write_dataframe_to_parquet_file(
    df: pd.DataFrame, file_path: Path, schema: Optional[pa.Schema] = None
):
    """
    Write the given dataframe to parquet file

    :param df: Pandas dataframe
    :param file_path: Path where the parquet file should be written
    :param schema: Arrow schema
    :return: NA
    """
    LOG.info(f'Writing df with shape {df.shape} to {file_path}')

    table = pa.Table.from_pandas(df, schema=schema, preserve_index=False)

    schema = schema or table.schema

    with pq.ParquetWriter(
        file_path, schema=schema, compression=PARQUET_COMPRESSION
    ) as pq_writer:
        pq_writer.write_table(table, row_group_size=PARQUET_ROW_GROUP_SIZE)
